#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA/ISAE
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, Iterable

import numpy as np

from fastoad.exceptions import FastUnexpectedKeywordArgument


class DynamicAttributeDict(dict):
    def __init__(self, *args, **kwargs):
        """
        A dictionary class where keys can also be used as attributes.

        The keys that can be used as attributes are defined using decorators
        :class:`AddKeyAttribute` or  :class:`SetKeyAttributes`.

        They can also be used as keyword arguments when instantiating this class.

        .. Note::

            Using this class as a dict is useful when instantiating another
            dict or a pandas DataFrame, or instantiating from them. Direct interaction
            with DynamicAttributeDict instance should be done through attributes.

        Example::

            >>> @AddKeyAttributes({"foo": 0.0, "bar": None, "baz": 42.0})
            ... class MyDict(DynamicAttributeDict):
            ...     pass
            ...

            >>> d = MyDict(foo=5, bar="aa")
            >>> d.foo
            5
            >>> d.bar
            'aa'
            >>> d.baz  # returns the default value
            42.0
            >>> d["foo"] = 10.0  # can still be used as a dict
            >>> d.foo  # but change are propagated to/from the matching attribute
            10.0
            >>> d.foo = np.nan  # setting None or numpy.nan returns to default value
            >>> d["foo"]
            0.0
            >>> d.foo # But the attribute will now return the default value
            0.0
            >>> d.bar = None  # If default value is None, setting None or numpy.nan deletes the key.
            >>> # d["bar"]  #would trigger a key error
            >>> d.bar # But the attribute will return None

        :param args: a dict-like object where all keys are contained in :attr:`attribute_keys`
        :param kwargs: argument keywords must be names contained in :attr:`attribute_keys`
        """

        if hasattr(self, "get_attribute_keys"):
            for key in kwargs:
                if key not in self.get_attribute_keys():
                    raise FastUnexpectedKeywordArgument(key)
        elif kwargs:
            # No defined dynamic attribute, any keyword argument is illegal
            raise FastUnexpectedKeywordArgument(list(kwargs.keys())[0])

        super().__init__(*args, **kwargs)

        if hasattr(self, "get_attribute_keys"):
            # Keys with None or Nan as value will be deleted so their default value
            # will be returned (see property definition in AddKeyAttribute).
            # We apply this behaviour even when instantiating from *args (i.e. with
            # a dict-like object)
            for key in self.get_attribute_keys():
                if key in self and (self[key] is None or _is_nan(self[key])):
                    del self[key]


class AddKeyAttribute:
    def __init__(self, attr_name, default_value=None, doc=None):
        """
        A decorator for a dict class that adds a property for accessing the matching dict item.

        The getter and the setter of the property are defined.
        Setting None or np.nan when setting the property will delete the dict key, so that
        next calls to the getter will return default_value.

        Calling AddKeyAttribute for an already defined key will redefine the default value.

        The "attribute_keys" property is created in decorated class for returning the list
        of attributes that have been defined by AddKeyAttribute or by
        :class:`AddKeyAttributes`.

        :param attr_name: the dict key that will be paired to a property
        :param default_value: the default value that will be returned if dict has not
                              the attr_name as key
        """
        self.attr_name = attr_name
        self.default_value = default_value
        self.doc = doc

    def __call__(self, decorated_dict: type):
        # Adds the property for defined key
        def _getter(self_):
            return self_.get(self.attr_name, self.default_value)

        def _setter(self_, value):
            if self.attr_name in self_ and (value is None or _is_nan(value)):
                # None or NaN means the default value, so we delete the dict key
                if self.default_value is None:
                    del self_[self.attr_name]
                else:
                    self_[self.attr_name] = self.default_value
            else:
                self_[self.attr_name] = value

        prop = property(_getter, _setter, doc=self.doc)
        setattr(decorated_dict, self.attr_name, prop)

        # Adds the method for getting the list of keys paired to attributes
        try:
            decorated_dict._attribute_keys.add(self.attr_name)
        except AttributeError:
            decorated_dict._attribute_keys = {self.attr_name}

        def get_attribute_keys(cls):
            """:return: list of attributes paired to dict key."""
            return cls._attribute_keys

        decorated_dict.get_attribute_keys = classmethod(get_attribute_keys)

        return decorated_dict


class AddKeyAttributes:
    def __init__(self, attribute_definition: Union[dict, Iterable[str]]):
        """
        A decorator for a dict class that adds properties for accessing the matching dict item.

        This class simply does several call of :class:`AddKeyAttribute`.

        :param attribute_definition: the list of keys that will be attributes. If it is
                                     a dictionary, the values are the associated default values.
                                     If it is a list or a set, default values will be None.
        """
        if isinstance(attribute_definition, dict):
            self.attribute_definition = attribute_definition
        else:
            self.attribute_definition = {attr_name: None for attr_name in attribute_definition}

    def __call__(self, decorated_dict: type):

        for attr_name, definition in self.attribute_definition.items():
            if isinstance(definition, dict):
                default_value = definition.get("default")
                doc = definition.get("doc")
            else:
                default_value = definition
                doc = None
            decorated_dict = AddKeyAttribute(attr_name, default_value, doc)(decorated_dict)

        return decorated_dict


def _is_nan(value):
    """Tells if value is numpy.nan in a robust way."""
    try:
        return np.all(np.isnan(value))
    except TypeError:
        return False  # if there has been a type error, then it is not NaN



# -*- coding: utf-8 -*-
"""
  Sellar openMDAO group
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import abc
from typing import Type

from openmdao.api import Group, IndepVarComp
from openmdao.api import NonlinearBlockGS

from .disc1 import Disc1
from .disc2 import Disc2
from .functions import Functions


class ISellarFactory(abc.ABC):
    """
    The interface for providing Sellar components. Is used by Sellar()
    """

    @staticmethod
    @abc.abstractmethod
    def create_disc1():
        """ Provides an instance for discipline 1 """

    @staticmethod
    @abc.abstractmethod
    def create_disc2():
        """ Provides an instance for discipline 2 """

    @staticmethod
    @abc.abstractmethod
    def create_functions():
        """ Provides an instance for functions """


class StandardSellarFactory(ISellarFactory):
    """
    Provides components through standard Python import
    """

    @staticmethod
    def create_disc1():
        return Disc1()

    @staticmethod
    def create_disc2():
        return Disc2()

    @staticmethod
    def create_functions():
        return Functions()


class Sellar(Group):
    """ An OpenMDAO base component to encapsulate Sellar MDA """

    def __init__(self, sellar_factory: Type[ISellarFactory] = StandardSellarFactory, **kwargs):
        """

        :param sellar_factory: will provide the components
                    (disciplines 1 and 2 + functions)
        :param thrift_client:
        :param kwargs:
        """
        super(Sellar, self).__init__(**kwargs)

        self._sellar_factory = sellar_factory
        self.nonlinear_solver = NonlinearBlockGS()
        self.nonlinear_solver.options["atol"] = 1.0e-10
        self.nonlinear_solver.options["rtol"] = 1.0e-10
        self.nonlinear_solver.options["maxiter"] = 10
        self.nonlinear_solver.options["err_on_non_converge"] = True
        self.nonlinear_solver.options["iprint"] = 1

    def setup(self):
        indeps = self.add_subsystem("indeps", IndepVarComp(), promotes=["*"])
        indeps.add_output("x", 2)
        indeps.add_output("z", [5, 2])
        self.add_subsystem(
            "Disc1", self._sellar_factory.create_disc1(), promotes=["x", "z", "y1", "y2"]
        )
        self.add_subsystem("Disc2", self._sellar_factory.create_disc2(), promotes=["z", "y1", "y2"])
        self.add_subsystem(
            "Functions",
            self._sellar_factory.create_functions(),
            promotes=["x", "z", "y1", "y2", "f", "g1", "g2"],
        )



"""
Tests module for standard_flight.py

These tests use the segment classes and the RubberEngine model, so they are
more integration tests than unit tests.
Therefore, obtained numerical results depend mainly on other classes, so this is
why almost no numerical check is done here (such checks will be done in the
non-regression tests).
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from os import mkdir
from shutil import rmtree

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from matplotlib.ticker import MultipleLocator
from numpy.testing import assert_allclose
from scipy.constants import knot, foot

from fastoad.base.flight_point import FlightPoint
from fastoad.constants import FlightPhase
from fastoad.models.propulsion.fuel_propulsion.base import FuelEngineSet
from fastoad.models.propulsion.fuel_propulsion.rubber_engine import RubberEngine
from ..standard_flight import StandardFlight
from ...flight.base import RangedFlight
from ...polar import Polar

RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)
    mkdir(RESULTS_FOLDER_PATH)


def print_dataframe(df, max_rows=20):
    """Utility for correctly printing results"""
    # Not used if all goes all well. Please keep it for future test setting/debugging.
    with pd.option_context(
        "display.max_rows", max_rows, "display.max_columns", None, "display.width", None
    ):
        print()
        print(df)


@pytest.fixture
def high_speed_polar() -> Polar:
    """Returns a dummy polar where max L/D ratio is about 16., around CL=0.5"""
    cl = np.arange(0.0, 1.5, 0.01)
    cd = 0.6e-1 * cl ** 2 + 0.016
    return Polar(cl, cd)


@pytest.fixture
def low_speed_polar() -> Polar:
    """Returns a dummy polar where max L/D ratio is around CL=0.5"""
    cl = np.arange(0.0, 2.0, 0.01)
    cd = 0.6e-1 * cl ** 2 + 0.03
    return Polar(cl, cd)


def plot_flight(flight_points, fig_filename):
    plt.figure(figsize=(12, 12))
    ax1 = plt.subplot(2, 1, 1)
    plt.plot(flight_points.ground_distance / 1000.0, flight_points.altitude / foot, "o-")
    plt.xlabel("distance [km]")
    plt.ylabel("altitude [ft]")
    ax1.xaxis.set_minor_locator(MultipleLocator(50))
    ax1.yaxis.set_minor_locator(MultipleLocator(500))
    plt.grid(which="major", color="k")
    plt.grid(which="minor")

    ax2 = plt.subplot(2, 1, 2)
    lines = []
    lines += plt.plot(
        flight_points.ground_distance / 1000.0, flight_points.true_airspeed, "b-", label="TAS [m/s]"
    )
    lines += plt.plot(
        flight_points.ground_distance / 1000.0,
        flight_points.equivalent_airspeed / knot,
        "g--",
        label="EAS [kt]",
    )
    plt.xlabel("distance [km]")
    plt.ylabel("speed")
    ax2.xaxis.set_minor_locator(MultipleLocator(50))
    ax2.yaxis.set_minor_locator(MultipleLocator(5))
    plt.grid(which="major", color="k")
    plt.grid(which="minor")

    plt.twinx(ax2)
    lines += plt.plot(
        flight_points.ground_distance / 1000.0, flight_points.mach, "r.-", label="Mach"
    )
    plt.ylabel("Mach")

    labels = [l.get_label() for l in lines]
    plt.legend(lines, labels, loc=0)

    plt.savefig(pth.join(RESULTS_FOLDER_PATH, fig_filename))
    plt.close()


def test_standard_flight_optimal_altitude(low_speed_polar, high_speed_polar, cleanup):

    engine = RubberEngine(5.0, 30.0, 1500.0, 1.0e5, 0.95, 10000.0)
    propulsion = FuelEngineSet(engine, 2)

    flight_calculator = StandardFlight(
        propulsion=propulsion,
        reference_area=120.0,
        low_speed_climb_polar=low_speed_polar,
        high_speed_polar=high_speed_polar,
        cruise_mach=0.78,
        thrust_rates={FlightPhase.CLIMB: 0.93, FlightPhase.DESCENT: 0.2},
        cruise_distance=4.0e6,
        time_step=None,
    )

    flight_points = flight_calculator.compute_from(
        FlightPoint(true_airspeed=150.0 * knot, altitude=100.0 * foot, mass=70000.0),
    )
    print_dataframe(flight_points)
    plot_flight(flight_points, "test_standard_flight_max_finesse.png")

    end_of_initial_climb = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.INITIAL_CLIMB.value].iloc[-1]
    )
    end_of_climb = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.CLIMB.value].iloc[-1]
    )
    end_of_cruise = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.CRUISE.value].iloc[-1]
    )
    end_of_descent = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.DESCENT.value].iloc[-1]
    )

    assert_allclose(end_of_initial_climb.altitude, 1500.0 * foot)
    assert_allclose(end_of_initial_climb.equivalent_airspeed, 250.0 * knot)
    assert_allclose(end_of_climb.mach, 0.78)
    assert end_of_climb.equivalent_airspeed <= 300.0 * knot
    assert_allclose(end_of_cruise.ground_distance - end_of_climb.ground_distance, 4.0e6)
    assert_allclose(end_of_cruise.mach, 0.78)
    assert_allclose(end_of_descent.altitude, 1500.0 * foot)
    assert_allclose(end_of_descent.equivalent_airspeed, 250.0 * knot)


def test_standard_flight_fixed_altitude(low_speed_polar, high_speed_polar, cleanup):
    # Wild version of an additional flight phase.
    # Start altitude is high enough to skip initial climb and first segment of climb.
    # Cruise altitude is low so that first segment of descent will also be skipped.

    engine = RubberEngine(5.0, 30.0, 1500.0, 1.0e5, 0.95, 10000.0)
    propulsion = FuelEngineSet(engine, 2)

    flight_calculator = StandardFlight(
        propulsion=propulsion,
        reference_area=120.0,
        low_speed_climb_polar=low_speed_polar,
        high_speed_polar=high_speed_polar,
        cruise_mach=0.78,
        thrust_rates={FlightPhase.CLIMB: 0.93, FlightPhase.DESCENT: 0.2},
        cruise_distance=4.0e6,
        climb_target_altitude=20000.0 * foot,
        descent_target_altitude=1000.0 * foot,
        time_step=None,
    )

    flight_points = flight_calculator.compute_from(
        FlightPoint(equivalent_airspeed=260.0 * knot, altitude=11000.0 * foot, mass=60000.0),
    )
    plot_flight(flight_points, "test_standard_flight_fixed_altitude.png")

    assert not any(flight_points.name == FlightPhase.INITIAL_CLIMB.value)
    end_of_climb = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.CLIMB.value].iloc[-1]
    )
    end_of_cruise = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.CRUISE.value].iloc[-1]
    )
    end_of_descent = FlightPoint(
        flight_points.loc[flight_points.name == FlightPhase.DESCENT.value].iloc[-1]
    )

    assert end_of_climb.mach < 0.78
    assert_allclose(end_of_climb.equivalent_airspeed, 300.0 * knot)
    assert_allclose(end_of_climb.altitude, 20000.0 * foot)
    assert_allclose(end_of_cruise.ground_distance - end_of_climb.ground_distance, 4.0e6)
    assert_allclose(end_of_descent.altitude, 1000.0 * foot)
    assert_allclose(end_of_descent.equivalent_airspeed, 250.0 * knot)


def test_ranged_flight(low_speed_polar, high_speed_polar, cleanup):

    engine = RubberEngine(5.0, 30.0, 1500.0, 1.0e5, 0.95, 10000.0)
    propulsion = FuelEngineSet(engine, 2)

    total_distance = 2.0e6
    flight_calculator = RangedFlight(
        StandardFlight(
            propulsion=propulsion,
            reference_area=120.0,
            low_speed_climb_polar=low_speed_polar,
            high_speed_polar=high_speed_polar,
            cruise_mach=0.78,
            thrust_rates={FlightPhase.CLIMB: 0.93, FlightPhase.DESCENT: 0.2},
        ),
        flight_distance=total_distance,
    )

    start = FlightPoint(
        true_airspeed=150.0 * knot, altitude=100.0 * foot, mass=70000.0, ground_distance=100000.0,
    )
    flight_points = flight_calculator.compute_from(start)

    plot_flight(flight_points, "test_ranged_flight.png")

    assert_allclose(
        flight_points.iloc[-1].ground_distance,
        total_distance + start.ground_distance,
        atol=flight_calculator.distance_accuracy,
    )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

from fastoad.openmdao.utils import get_problem_after_setup


def set_all_input_defaults(model: om.Group):
    """
    Sets needed defaults to avoid error messages about ambiguous variable definitions.

    :param model: the model where defaults will be added in place.
    """
    # FIXME: This solution is a patch to get compatible with OpenMDAO 3.2+ but it not really
    #  satisfactory because it does not help in case of custom models using existing variables
    #  with different units.
    #  Maybe a future evolution of OpenMDAO will open the way to a better solution...

    variables = {
        "data:geometry:wing:sweep_25": dict(val=np.nan, units="deg"),
        "data:TLAR:range": dict(val=np.nan, units="NM"),
    }

    # Defaults must be added only if variable is present. Otherwise, an error is raised at setup().
    # Therefore, a setup() is done and if OpenMDAO complains about missing default and if that
    # default is available, it is added.
    ok = False
    while not ok and len(variables) > 0:  # Stops if no error or if no default definition is left.
        try:
            get_problem_after_setup(om.Problem(model))
            ok = True
        except RuntimeError as exc:
            var_default_has_been_set = False
            for var_name, metadata in variables.items():
                if var_name in exc.args[0]:
                    model.set_input_defaults(var_name, **metadata)
                    var_default_has_been_set = True
                    break
            if var_default_has_been_set:
                del variables[var_name]



"""
    Estimation of wing geometry
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.models.geometry.geom_components.wing.components.compute_b_50 import ComputeB50
from fastoad.models.geometry.geom_components.wing.components.compute_cl_alpha import ComputeCLalpha
from fastoad.models.geometry.geom_components.wing.components.compute_l1_l4 import ComputeL1AndL4Wing
from fastoad.models.geometry.geom_components.wing.components.compute_l2_l3 import ComputeL2AndL3Wing
from fastoad.models.geometry.geom_components.wing.components.compute_mac_wing import ComputeMACWing
from fastoad.models.geometry.geom_components.wing.components.compute_mfw import ComputeMFW
from fastoad.models.geometry.geom_components.wing.components.compute_sweep_wing import (
    ComputeSweepWing,
)
from fastoad.models.geometry.geom_components.wing.components.compute_toc_wing import ComputeToCWing
from fastoad.models.geometry.geom_components.wing.components.compute_wet_area_wing import (
    ComputeWetAreaWing,
)
from fastoad.models.geometry.geom_components.wing.components.compute_x_wing import ComputeXWing
from fastoad.models.geometry.geom_components.wing.components.compute_y_wing import ComputeYWing

from openmdao.api import Group


class ComputeWingGeometry(Group):
    # TODO: Document equations. Cite sources
    """ Wing geometry estimation """

    def setup(self):
        self.add_subsystem("y_wing", ComputeYWing(), promotes=["*"])
        self.add_subsystem("l14_wing", ComputeL1AndL4Wing(), promotes=["*"])
        self.add_subsystem("l2l3_wing", ComputeL2AndL3Wing(), promotes=["*"])
        self.add_subsystem("x_wing", ComputeXWing(), promotes=["*"])
        self.add_subsystem("mac_wing", ComputeMACWing(), promotes=["*"])
        self.add_subsystem("b50_wing", ComputeB50(), promotes=["*"])
        self.add_subsystem("sweep_wing", ComputeSweepWing(), promotes=["*"])
        self.add_subsystem("toc_wing", ComputeToCWing(), promotes=["*"])
        self.add_subsystem("wetarea_wing", ComputeWetAreaWing(), promotes=["*"])
        self.add_subsystem("clapha_wing", ComputeCLalpha(), promotes=["*"])
        self.add_subsystem("mfw", ComputeMFW(), promotes=["*"])



"""
Test module for mass breakdown functions
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=redefined-outer-name  # needed for pytest fixtures
import os.path as pth

import openmdao.api as om
import pytest
from fastoad.io import VariableIO

from tests.testing_utilities import run_system
from ..a_airframe import (
    EmpennageWeight,
    FlightControlsWeight,
    FuselageWeight,
    PaintWeight,
    PylonsWeight,
    WingWeight,
    LandingGearWeight,
)
from ..b_propulsion import FuelLinesWeight, UnconsumablesWeight, EngineWeight
from ..c_systems import (
    FixedOperationalSystemsWeight,
    FlightKitWeight,
    LifeSupportSystemsWeight,
    NavigationSystemsWeight,
    PowerSystemsWeight,
    TransmissionSystemsWeight,
)
from ..cs25 import Loads
from ..d_furniture import (
    CargoConfigurationWeight,
    PassengerSeatsWeight,
    FoodWaterWeight,
    ToiletsWeight,
    SecurityKitWeight,
)
from ..e_crew import CrewWeight
from ..mass_breakdown import MassBreakdown, OperatingWeightEmpty
from ..payload import ComputePayload


def get_indep_var_comp(var_names):
    """ Reads required input data and returns an IndepVarcomp() instance"""
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "mass_breakdown_inputs.xml"))
    reader.path_separator = ":"
    ivc = reader.read(only=var_names).to_ivc()
    return ivc


def test_compute_payload():
    ivc = om.IndepVarComp()
    ivc.add_output("data:TLAR:NPAX", val=150)
    problem = run_system(ComputePayload(), ivc)
    assert problem["data:weight:aircraft:payload"] == pytest.approx(13608.0, abs=0.1)
    assert problem["data:weight:aircraft:max_payload"] == pytest.approx(19608.0, abs=0.1)

    ivc = om.IndepVarComp()
    ivc.add_output("data:TLAR:NPAX", val=150)
    ivc.add_output(
        "settings:weight:aircraft:payload:design_mass_per_passenger", val=1.0, units="kg"
    )
    ivc.add_output("settings:weight:aircraft:payload:max_mass_per_passenger", val=2.0, units="kg")
    problem = run_system(ComputePayload(), ivc)
    assert problem["data:weight:aircraft:payload"] == pytest.approx(150.0, abs=0.1)
    assert problem["data:weight:aircraft:max_payload"] == pytest.approx(300.0, abs=0.1)


def test_compute_loads():
    """ Tests computation of sizing loads """
    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:span",
        "data:weight:aircraft:MZFW",
        "data:weight:aircraft:MFW",
        "data:weight:aircraft:MTOW",
        "data:aerodynamics:aircraft:cruise:CL_alpha",
        "data:load_case:lc1:U_gust",
        "data:load_case:lc1:altitude",
        "data:load_case:lc1:Vc_EAS",
        "data:load_case:lc2:U_gust",
        "data:load_case:lc2:altitude",
        "data:load_case:lc2:Vc_EAS",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(Loads(), ivc)

    n1m1 = problem["data:mission:sizing:cs25:sizing_load_1"]
    n2m2 = problem["data:mission:sizing:cs25:sizing_load_2"]
    assert n1m1 == pytest.approx(240968, abs=10)
    assert n2m2 == pytest.approx(254130, abs=10)


def test_compute_wing_weight():
    """ Tests wing weight computation from sample XML data """
    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:span",
        "data:geometry:wing:root:thickness_ratio",
        "data:geometry:wing:kink:thickness_ratio",
        "data:geometry:wing:tip:thickness_ratio",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:outer_area",
        "data:weight:aircraft:MTOW",
        "data:weight:aircraft:MLW",
        "tuning:weight:airframe:wing:mass:k",
        "tuning:weight:airframe:wing:mass:offset",
        "tuning:weight:airframe:wing:bending_sizing:mass:k",
        "tuning:weight:airframe:wing:bending_sizing:mass:offset",
        "tuning:weight:airframe:wing:shear_sizing:mass:k",
        "tuning:weight:airframe:wing:shear_sizing:mass:offset",
        "tuning:weight:airframe:wing:ribs:mass:k",
        "tuning:weight:airframe:wing:ribs:mass:offset",
        "tuning:weight:airframe:wing:reinforcements:mass:k",
        "tuning:weight:airframe:wing:reinforcements:mass:offset",
        "tuning:weight:airframe:wing:secondary_parts:mass:k",
        "tuning:weight:airframe:wing:secondary_parts:mass:offset",
        "settings:weight:airframe:wing:mass:k_voil",
        "settings:weight:airframe:wing:mass:k_mvo",
    ]

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:mission:sizing:cs25:sizing_load_1", 241000, units="kg")
    ivc.add_output("data:mission:sizing:cs25:sizing_load_2", 250000, units="kg")

    problem = run_system(WingWeight(), ivc)

    val = problem["data:weight:airframe:wing:mass"]
    assert val == pytest.approx(7681, abs=1)


def test_compute_fuselage_weight():
    """ Tests fuselage weight computation from sample XML data """
    input_list = [
        "data:geometry:fuselage:wetted_area",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "tuning:weight:airframe:fuselage:mass:k",
        "tuning:weight:airframe:fuselage:mass:offset",
        "settings:weight:airframe:fuselage:mass:k_lg",
        "settings:weight:airframe:fuselage:mass:k_fus",
    ]

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:mission:sizing:cs25:sizing_load_1", 241000, units="kg")

    problem = run_system(FuselageWeight(), ivc)

    val = problem["data:weight:airframe:fuselage:mass"]
    assert val == pytest.approx(8828, abs=1)


def test_compute_empennage_weight():
    """ Tests empennage weight computation from sample XML data """
    input_list = [
        "data:geometry:has_T_tail",
        "data:geometry:horizontal_tail:area",
        "data:geometry:vertical_tail:area",
        "data:geometry:propulsion:layout",
        "tuning:weight:airframe:horizontal_tail:mass:k",
        "tuning:weight:airframe:horizontal_tail:mass:offset",
        "tuning:weight:airframe:vertical_tail:mass:k",
        "tuning:weight:airframe:vertical_tail:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(EmpennageWeight(), ivc)
    val1 = problem["data:weight:airframe:horizontal_tail:mass"]
    val2 = problem["data:weight:airframe:vertical_tail:mass"]
    assert val1 == pytest.approx(754, abs=1)
    assert val2 == pytest.approx(515, abs=1)


def test_compute_flight_controls_weight():
    """ Tests flight controls weight computation from sample XML data """
    input_list = [
        "data:geometry:fuselage:length",
        "data:geometry:wing:b_50",
        "tuning:weight:airframe:flight_controls:mass:k",
        "tuning:weight:airframe:flight_controls:mass:offset",
        "settings:weight:airframe:flight_controls:mass:k_fc",
    ]
    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:mission:sizing:cs25:sizing_load_1", 241000, units="kg")
    ivc.add_output("data:mission:sizing:cs25:sizing_load_2", 250000, units="kg")
    problem = run_system(FlightControlsWeight(), ivc)

    val = problem["data:weight:airframe:flight_controls:mass"]
    assert val == pytest.approx(716, abs=1)


def test_compute_landing_gear_weight():
    """ Tests landing gear weight computation from sample XML data """
    input_list = [
        "data:weight:aircraft:MTOW",
        "tuning:weight:airframe:landing_gear:mass:k",
        "tuning:weight:airframe:landing_gear:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(LandingGearWeight(), ivc)

    val1 = problem["data:weight:airframe:landing_gear:main:mass"]
    val2 = problem["data:weight:airframe:landing_gear:front:mass"]
    assert val1 == pytest.approx(2144, abs=1)
    assert val2 == pytest.approx(379, abs=1)


def test_compute_pylons_weight():
    """ Tests pylons weight computation from sample XML data """
    input_list = [
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:propulsion:engine:count",
        "data:geometry:propulsion:layout",
        "tuning:weight:airframe:pylon:mass:k",
        "tuning:weight:airframe:pylon:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:weight:propulsion:engine:mass", 7161.33, units="kg")
    problem = run_system(PylonsWeight(), ivc)

    val = problem["data:weight:airframe:pylon:mass"]
    assert val == pytest.approx(1212, abs=1)


def test_compute_paint_weight():
    """ Tests paint weight computation from sample XML data """
    input_list = [
        "data:geometry:aircraft:wetted_area",
        "tuning:weight:airframe:paint:mass:k",
        "tuning:weight:airframe:paint:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(PaintWeight(), ivc)

    val = problem["data:weight:airframe:paint:mass"]
    assert val == pytest.approx(141.1, abs=0.1)


def test_compute_engine_weight():
    """ Tests engine weight computation from sample XML data """
    input_list = [
        "data:propulsion:MTO_thrust",
        "data:geometry:propulsion:engine:count",
        "tuning:weight:propulsion:engine:mass:k",
        "tuning:weight:propulsion:engine:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(EngineWeight(), ivc)

    val = problem["data:weight:propulsion:engine:mass"]
    assert val == pytest.approx(7161, abs=1)


def test_compute_fuel_lines_weight():
    """ Tests fuel lines weight computation from sample XML data """
    input_list = [
        "data:geometry:wing:b_50",
        "data:weight:aircraft:MFW",
        "tuning:weight:propulsion:fuel_lines:mass:k",
        "tuning:weight:propulsion:fuel_lines:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:weight:propulsion:engine:mass", 7161.33, units="kg")
    problem = run_system(FuelLinesWeight(), ivc)

    val = problem["data:weight:propulsion:fuel_lines:mass"]
    assert val == pytest.approx(457, abs=1)


def test_compute_unconsumables_weight():
    """ Tests "unconsumables" weight computation from sample XML data """
    input_list = [
        "data:geometry:propulsion:engine:count",
        "data:weight:aircraft:MFW",
        "tuning:weight:propulsion:unconsumables:mass:k",
        "tuning:weight:propulsion:unconsumables:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(UnconsumablesWeight(), ivc)

    val = problem["data:weight:propulsion:unconsumables:mass"]
    assert val == pytest.approx(122, abs=1)


def test_compute_power_systems_weight():
    """ Tests power systems weight computation from sample XML data """
    input_list = [
        "data:weight:aircraft:MTOW",
        "tuning:weight:systems:power:auxiliary_power_unit:mass:k",
        "tuning:weight:systems:power:auxiliary_power_unit:mass:offset",
        "tuning:weight:systems:power:electric_systems:mass:k",
        "tuning:weight:systems:power:electric_systems:mass:offset",
        "tuning:weight:systems:power:hydraulic_systems:mass:k",
        "tuning:weight:systems:power:hydraulic_systems:mass:offset",
        "settings:weight:systems:power:mass:k_elec",
    ]
    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:geometry:cabin:NPAX1", 150)
    ivc.add_output("data:weight:airframe:flight_controls:mass", 700, units="kg")
    problem = run_system(PowerSystemsWeight(), ivc)

    val1 = problem["data:weight:systems:power:auxiliary_power_unit:mass"]
    val2 = problem["data:weight:systems:power:electric_systems:mass"]
    val3 = problem["data:weight:systems:power:hydraulic_systems:mass"]
    assert val1 == pytest.approx(279, abs=1)
    assert val2 == pytest.approx(1297, abs=1)
    assert val3 == pytest.approx(747, abs=1)


def test_compute_life_support_systems_weight():
    """ Tests life support systems weight computation from sample XML data """
    input_list = [
        "data:TLAR:range",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:cabin:length",
        "data:geometry:wing:sweep_0",
        "data:geometry:propulsion:nacelle:diameter",
        "data:geometry:propulsion:engine:count",
        "data:geometry:cabin:crew_count:technical",
        "data:geometry:cabin:crew_count:commercial",
        "data:geometry:wing:span",
        "tuning:weight:systems:life_support:insulation:mass:k",
        "tuning:weight:systems:life_support:insulation:mass:offset",
        "tuning:weight:systems:life_support:air_conditioning:mass:k",
        "tuning:weight:systems:life_support:air_conditioning:mass:offset",
        "tuning:weight:systems:life_support:de-icing:mass:k",
        "tuning:weight:systems:life_support:de-icing:mass:offset",
        "tuning:weight:systems:life_support:cabin_lighting:mass:k",
        "tuning:weight:systems:life_support:cabin_lighting:mass:offset",
        "tuning:weight:systems:life_support:seats_crew_accommodation:mass:k",
        "tuning:weight:systems:life_support:seats_crew_accommodation:mass:offset",
        "tuning:weight:systems:life_support:oxygen:mass:k",
        "tuning:weight:systems:life_support:oxygen:mass:offset",
        "tuning:weight:systems:life_support:safety_equipment:mass:k",
        "tuning:weight:systems:life_support:safety_equipment:mass:offset",
    ]
    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:geometry:cabin:NPAX1", 150)
    ivc.add_output("data:weight:propulsion:engine:mass", 7161.33, units="kg")
    problem = run_system(LifeSupportSystemsWeight(), ivc)

    val1 = problem["data:weight:systems:life_support:insulation:mass"]
    val2 = problem["data:weight:systems:life_support:air_conditioning:mass"]
    val3 = problem["data:weight:systems:life_support:de-icing:mass"]
    val4 = problem["data:weight:systems:life_support:cabin_lighting:mass"]
    val5 = problem["data:weight:systems:life_support:seats_crew_accommodation:mass"]
    val6 = problem["data:weight:systems:life_support:oxygen:mass"]
    val7 = problem["data:weight:systems:life_support:safety_equipment:mass"]
    assert val1 == pytest.approx(2226, abs=1)
    assert val2 == pytest.approx(920, abs=1)
    assert val3 == pytest.approx(154, abs=1)
    assert val4 == pytest.approx(168, abs=1)
    assert val5 == pytest.approx(126, abs=1)
    assert val6 == pytest.approx(275, abs=1)
    assert val7 == pytest.approx(416, abs=1)


def test_compute_navigation_systems_weight():
    """ Tests navigation systems weight computation from sample XML data """
    input_list = [
        "data:geometry:fuselage:length",
        "data:geometry:wing:b_50",
        "tuning:weight:systems:navigation:mass:k",
        "tuning:weight:systems:navigation:mass:offset",
    ]
    component = NavigationSystemsWeight()

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="km")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:navigation:mass"] == pytest.approx(193, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:navigation:mass"] == pytest.approx(493, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 4000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:navigation:mass"] == pytest.approx(743, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 5000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:navigation:mass"] == pytest.approx(843, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 8000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:navigation:mass"] == pytest.approx(843, abs=1)


def test_compute_transmissions_systems_weight():
    """ Tests transmissions weight computation from sample XML data """
    input_list = [
        "tuning:weight:systems:transmission:mass:k",
        "tuning:weight:systems:transmission:mass:offset",
    ]

    component = TransmissionSystemsWeight()

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="km")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:transmission:mass"] == pytest.approx(100, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:transmission:mass"] == pytest.approx(200, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 4000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:transmission:mass"] == pytest.approx(250, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 5000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:transmission:mass"] == pytest.approx(350, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 8000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:transmission:mass"] == pytest.approx(350, abs=1)


def test_compute_fixed_operational_systems_weight():
    """
    Tests fixed operational systems weight computation from sample XML data
    """
    input_list = [
        "data:geometry:fuselage:front_length",
        "data:geometry:fuselage:rear_length",
        "data:geometry:fuselage:length",
        "data:geometry:cabin:seats:economical:count_by_row",
        "data:geometry:wing:root:chord",
        "data:geometry:cabin:containers:count_by_row",
        "tuning:weight:systems:operational:mass:k",
        "tuning:weight:systems:operational:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(FixedOperationalSystemsWeight(), ivc)

    val1 = problem["data:weight:systems:operational:radar:mass"]
    val2 = problem["data:weight:systems:operational:cargo_hold:mass"]
    assert val1 == pytest.approx(100, abs=1)
    assert val2 == pytest.approx(277, abs=1)


def test_compute_flight_kit_weight():
    """ Tests flight kit weight computation from sample XML data """
    input_list = [
        "tuning:weight:systems:flight_kit:mass:k",
        "tuning:weight:systems:flight_kit:mass:offset",
    ]
    component = FlightKitWeight()

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="km")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:flight_kit:mass"] == pytest.approx(10, abs=1)

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:TLAR:range", 2000.0, units="NM")
    problem = run_system(component, ivc)
    assert problem["data:weight:systems:flight_kit:mass"] == pytest.approx(45, abs=1)


def test_compute_cargo_configuration_weight():
    """ Tests cargo configuration weight computation from sample XML data """
    input_list = [
        "data:geometry:cabin:containers:count",
        "data:geometry:cabin:pallet_count",
        "data:geometry:cabin:seats:economical:count_by_row",
        "tuning:weight:furniture:cargo_configuration:mass:k",
        "tuning:weight:furniture:cargo_configuration:mass:offset",
    ]
    component = CargoConfigurationWeight()

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:geometry:cabin:NPAX1", 150)
    problem = run_system(component, ivc)
    val = problem["data:weight:furniture:cargo_configuration:mass"]
    assert val == pytest.approx(39.3, abs=0.1)


def test_compute_passenger_seats_weight():
    """ Tests passenger seats weight computation from sample XML data """
    input_list = [
        "data:TLAR:range",
        "data:TLAR:NPAX",
        "tuning:weight:furniture:passenger_seats:mass:k",
        "tuning:weight:furniture:passenger_seats:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    component = PassengerSeatsWeight()

    problem = run_system(component, ivc)
    val = problem["data:weight:furniture:passenger_seats:mass"]
    assert val == pytest.approx(1500, abs=1)


def test_compute_food_water_weight():
    """ Tests food water weight computation from sample XML data """
    input_list = [
        "data:TLAR:range",
        "data:TLAR:NPAX",
        "tuning:weight:furniture:food_water:mass:k",
        "tuning:weight:furniture:food_water:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    component = FoodWaterWeight()

    problem = run_system(component, ivc)
    val = problem["data:weight:furniture:food_water:mass"]
    assert val == pytest.approx(1312, abs=1)


def test_compute_security_kit_weight():
    """ Tests security kit weight computation from sample XML data """
    input_list = [
        "data:TLAR:range",
        "data:TLAR:NPAX",
        "tuning:weight:furniture:security_kit:mass:k",
        "tuning:weight:furniture:security_kit:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    component = SecurityKitWeight()

    problem = run_system(component, ivc)
    val = problem["data:weight:furniture:security_kit:mass"]
    assert val == pytest.approx(225, abs=1)


def test_compute_toilets_weight():
    """ Tests toilets weight computation from sample XML data """
    input_list = [
        "data:TLAR:range",
        "data:TLAR:NPAX",
        "tuning:weight:furniture:toilets:mass:k",
        "tuning:weight:furniture:toilets:mass:offset",
    ]

    ivc = get_indep_var_comp(input_list)
    component = ToiletsWeight()

    problem = run_system(component, ivc)
    val = problem["data:weight:furniture:toilets:mass"]
    assert val == pytest.approx(75, abs=0.1)


def test_compute_crew_weight():
    """ Tests crew weight computation from sample XML data """
    input_list = [
        "data:geometry:cabin:crew_count:technical",
        "data:geometry:cabin:crew_count:commercial",
    ]
    ivc = get_indep_var_comp(input_list)
    problem = run_system(CrewWeight(), ivc)

    val = problem["data:weight:crew:mass"]
    assert val == pytest.approx(470, abs=1)


def test_evaluate_oew():
    """
    Tests a simple evaluation of Operating Empty Weight from sample XML data.
    """
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "mass_breakdown_inputs.xml"))
    reader.path_separator = ":"
    input_vars = reader.read().to_ivc()

    mass_computation = run_system(OperatingWeightEmpty(), input_vars)

    oew = mass_computation["data:weight:aircraft:OWE"]
    assert oew == pytest.approx(41591, abs=1)


def test_loop_compute_oew():
    """
    Tests a weight computation loop matching the max payload criterion.
    """
    # With payload from npax
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "mass_breakdown_inputs.xml"))
    reader.path_separator = ":"
    input_vars = reader.read(
        ignore=[
            "data:weight:aircraft:MLW",
            "data:weight:aircraft:MZFW",
            "data:weight:aircraft:max_payload",
        ]
    ).to_ivc()
    mass_computation = run_system(MassBreakdown(), input_vars)
    oew = mass_computation["data:weight:aircraft:OWE"]
    assert oew == pytest.approx(41591, abs=1)

    # with payload as input
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "mass_breakdown_inputs.xml"))
    reader.path_separator = ":"
    input_vars = reader.read(
        ignore=["data:weight:aircraft:MLW", "data:weight:aircraft:MZFW",]
    ).to_ivc()
    mass_computation = run_system(MassBreakdown(payload_from_npax=False), input_vars)
    oew = mass_computation["data:weight:aircraft:OWE"]
    assert oew == pytest.approx(42060, abs=1)



"""
This module launches XFOIL computations
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import os.path as pth
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
from importlib_resources import path
from openmdao.components.external_code_comp import ExternalCodeComp
from openmdao.utils.file_wrap import InputFileGenerator

from fastoad.models.aerodynamics.external.xfoil import xfoil699
from fastoad.models.geometry.profiles.get_profile import get_profile
from fastoad.utils.resource_management.copy import copy_resource
from . import resources

OPTION_RESULT_POLAR_FILENAME = "result_polar_filename"
OPTION_RESULT_FOLDER_PATH = "result_folder_path"
OPTION_PROFILE_NAME = "profile_name"
OPTION_XFOIL_EXE_PATH = "xfoil_exe_path"
OPTION_ALPHA_START = "alpha_start"
OPTION_ALPHA_END = "alpha_end"
OPTION_ITER_LIMIT = "iter_limit"

DEFAULT_2D_CL_MAX = 1.9

_INPUT_FILE_NAME = "polar_session.txt"
_STDOUT_FILE_NAME = "polar_calc.log"
_STDERR_FILE_NAME = "polar_calc.err"
_TMP_PROFILE_FILE_NAME = "in"  # as short as possible to avoid problems of path length
_TMP_RESULT_FILE_NAME = "out"  # as short as possible to avoid problems of path length

XFOIL_EXE_NAME = "xfoil.exe"  # name of embedded XFoil executable
DEFAULT_PROFILE_FILENAME = "BACJ.txt"

_LOGGER = logging.getLogger(__name__)

_XFOIL_PATH_LIMIT = 64


class XfoilPolar(ExternalCodeComp):
    """
    Runs a polar computation with XFOIL and returns the 2D max lift coefficient
    """

    _xfoil_output_names = ["alpha", "CL", "CD", "CDp", "CM", "Top_Xtr", "Bot_Xtr"]
    """Column names in XFOIL polar result"""

    def initialize(self):
        self.options.declare(OPTION_XFOIL_EXE_PATH, default="", types=str, allow_none=True)
        self.options.declare(OPTION_PROFILE_NAME, default="BACJ.txt", types=str)
        self.options.declare(OPTION_RESULT_FOLDER_PATH, default="", types=str)
        self.options.declare(OPTION_RESULT_POLAR_FILENAME, default="polar_result.txt", types=str)
        self.options.declare(OPTION_ALPHA_START, default=0.0, types=float)
        self.options.declare(OPTION_ALPHA_END, default=30.0, types=float)
        self.options.declare(OPTION_ITER_LIMIT, default=500, types=int)

    def setup(self):

        self.add_input("xfoil:reynolds", val=np.nan)
        self.add_input("xfoil:mach", val=np.nan)
        self.add_input("data:geometry:wing:thickness_ratio", val=np.nan)

        self.add_output("xfoil:CL_max_2D")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):

        # Create result folder first (if it must fail, let it fail as soon as possible)
        result_folder_path = self.options[OPTION_RESULT_FOLDER_PATH]
        if result_folder_path != "":
            os.makedirs(result_folder_path, exist_ok=True)

        # Get inputs
        reynolds = inputs["xfoil:reynolds"]
        mach = inputs["xfoil:mach"]
        thickness_ratio = inputs["data:geometry:wing:thickness_ratio"]

        # Pre-processing (populating temp directory) -----------------------------------------------
        # XFoil exe
        tmp_directory = self._create_tmp_directory()
        if self.options[OPTION_XFOIL_EXE_PATH]:
            # if a path for Xfoil has been provided, simply use it
            self.options["command"] = [self.options[OPTION_XFOIL_EXE_PATH]]
        else:
            # otherwise, copy the embedded resource in tmp dir
            copy_resource(xfoil699, XFOIL_EXE_NAME, tmp_directory.name)
            self.options["command"] = [pth.join(tmp_directory.name, XFOIL_EXE_NAME)]

        # I/O files
        self.stdin = pth.join(tmp_directory.name, _INPUT_FILE_NAME)
        self.stdout = pth.join(tmp_directory.name, _STDOUT_FILE_NAME)
        self.stderr = pth.join(tmp_directory.name, _STDERR_FILE_NAME)

        # profile file
        tmp_profile_file_path = pth.join(tmp_directory.name, _TMP_PROFILE_FILE_NAME)
        profile = get_profile(
            file_name=self.options[OPTION_PROFILE_NAME], thickness_ratio=thickness_ratio
        )
        np.savetxt(
            tmp_profile_file_path,
            profile.to_numpy(),
            fmt="%.15f",
            delimiter=" ",
            header="Wing",
            comments="",
        )

        # standard input file
        tmp_result_file_path = pth.join(tmp_directory.name, _TMP_RESULT_FILE_NAME)
        parser = InputFileGenerator()
        with path(resources, _INPUT_FILE_NAME) as input_template_path:
            parser.set_template_file(input_template_path)
            parser.set_generated_file(self.stdin)

            # Fills numeric values
            parser.mark_anchor("RE")
            parser.transfer_var(float(reynolds), 1, 1)
            parser.mark_anchor("M")
            parser.transfer_var(float(mach), 1, 1)
            parser.mark_anchor("ITER")
            parser.transfer_var(self.options[OPTION_ITER_LIMIT], 1, 1)
            parser.mark_anchor("ASEQ")
            parser.transfer_var(self.options[OPTION_ALPHA_START], 1, 1)
            parser.transfer_var(self.options[OPTION_ALPHA_END], 2, 1)

            # Fills string values
            # If a provide path contains the string that is used as next anchor, the process
            # will fail. Doing these replacements at the end prevent this to happen.
            parser.reset_anchor()
            parser.mark_anchor("LOAD")
            parser.transfer_var(tmp_profile_file_path, 1, 1)
            parser.mark_anchor("PACC", -2)
            parser.transfer_var(tmp_result_file_path, 1, 1)

            parser.generate()

        # Run XFOIL --------------------------------------------------------------------------------
        self.options["external_input_files"] = [self.stdin, tmp_profile_file_path]
        self.options["external_output_files"] = [tmp_result_file_path]
        super().compute(inputs, outputs)

        # Post-processing --------------------------------------------------------------------------
        result_array = self._read_polar(tmp_result_file_path)
        outputs["xfoil:CL_max_2D"] = self._get_max_cl(result_array["alpha"], result_array["CL"])

        # Getting output files if needed
        if self.options[OPTION_RESULT_FOLDER_PATH]:
            if pth.exists(tmp_result_file_path):
                polar_file_path = pth.join(
                    result_folder_path, self.options[OPTION_RESULT_POLAR_FILENAME]
                )
                shutil.move(tmp_result_file_path, polar_file_path)

            if pth.exists(self.stdin):
                stdin_file_path = pth.join(result_folder_path, _INPUT_FILE_NAME)
                shutil.move(self.stdin, stdin_file_path)

            if pth.exists(self.stdout):
                stdout_file_path = pth.join(result_folder_path, _STDOUT_FILE_NAME)
                shutil.move(self.stdout, stdout_file_path)

            if pth.exists(self.stderr):
                stderr_file_path = pth.join(result_folder_path, _STDERR_FILE_NAME)
                shutil.move(self.stderr, stderr_file_path)

        tmp_directory.cleanup()

    @staticmethod
    def _read_polar(xfoil_result_file_path: str) -> np.ndarray:
        """
        :param xfoil_result_file_path:
        :return: numpy array with XFoil polar results
        """
        if os.path.isfile(xfoil_result_file_path):
            dtypes = [(name, "f8") for name in XfoilPolar._xfoil_output_names]
            result_array = np.genfromtxt(xfoil_result_file_path, skip_header=12, dtype=dtypes)
            return result_array

        _LOGGER.error("XFOIL results file not found")
        return np.array([])

    @staticmethod
    def _get_max_cl(alpha: np.ndarray, lift_coeff: np.ndarray) -> float:
        """

        :param alpha:
        :param lift_coeff: CL
        :return: max CL if enough alpha computed, or default value otherwise
        """
        if len(alpha) > 0 and max(alpha) >= 5.0:
            return max(lift_coeff)

        _LOGGER.warning("2D CL max not found. Using default value (%s)", DEFAULT_2D_CL_MAX)
        return DEFAULT_2D_CL_MAX

    @staticmethod
    def _create_tmp_directory() -> TemporaryDirectory:
        # Dev Note: XFOIL fails if length of provided file path exceeds 64 characters.
        #           Changing working directory to the tmp dir would allow to just provide file name,
        #           but it is not really safe (at least, it does mess with the coverage report).
        #           Then the point is to get a tmp directory with a short path.
        #           On Windows, the default (user-dependent) tmp dir can exceed the limit.
        #           Therefore, as a second choice, tmp dir is created as close of user home
        #           directory as possible.

        tmp_base_candidates = [None, pth.join(str(Path.home()), ".fast")]

        tmp_candidates = []
        for tmp_base_path in tmp_base_candidates:
            if tmp_base_path is not None:
                os.makedirs(tmp_base_path, exist_ok=True)
            tmp_directory = TemporaryDirectory(dir=tmp_base_path)
            tmp_candidates.append(tmp_directory.name)
            tmp_profile_file_path = pth.join(tmp_directory.name, _TMP_PROFILE_FILE_NAME)
            tmp_result_file_path = pth.join(tmp_directory.name, _TMP_RESULT_FILE_NAME)

            if max(len(tmp_profile_file_path), len(tmp_result_file_path)) <= _XFOIL_PATH_LIMIT:
                # tmp_directory is OK. Stop there
                break
            # tmp_directory has a too long path. Erase and continue...
            tmp_directory.cleanup()

        if max(len(tmp_profile_file_path), len(tmp_result_file_path)) > _XFOIL_PATH_LIMIT:
            raise IOError(
                "Could not create a tmp directory where file path will respect XFOIL "
                "limitation (%i): tried %s" % (_XFOIL_PATH_LIMIT, tmp_candidates)
            )

        return tmp_directory



"""
Basis for registering and retrieving services
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import re
from typing import List, Tuple, Set, Union, Optional, Any

import pelix
from pelix.framework import FrameworkFactory, Framework, Bundle, BundleContext
from pelix.internals.registry import ServiceReference
from pelix.ipopo.constants import SERVICE_IPOPO, use_ipopo
from pelix.ipopo.decorators import ComponentFactory, Provides, Property

from .exceptions import FastDuplicateFactoryError

_LOGGER = logging.getLogger(__name__)
"""Logger for this module"""


class BundleLoader:
    """
    Helper class for loading Pelix bundles.
    """

    def __init__(self):
        """
        Constructor
        """
        self._framework = None

    @property
    def framework(self) -> Framework:
        """
        The currently running Pelix framework instance, or a new one if none is running (anyway,
        the framework instance will continue after deletion of this BundleLoader instance)
        """
        if FrameworkFactory.is_framework_running():
            self._framework = FrameworkFactory.get_framework()
            # Do not use self.context, because it will call back this.
            context = self._framework.get_bundle_context()
            if not context.get_service_reference(SERVICE_IPOPO):
                self._framework.install_bundle(SERVICE_IPOPO)
        else:
            self._framework = pelix.framework.create_framework(())
            self._framework.install_bundle(SERVICE_IPOPO)
            self._framework.start()
        return self._framework

    @property
    def context(self) -> BundleContext:
        """
        BundleContext instance of running Pelix framework
        """
        return self.framework.get_bundle_context()

    def install_packages(self, folder_path: str) -> Tuple[Set[Bundle], Set[str]]:
        """
        Installs bundles found in *folder_path*.

        Bundles that contain factories that are programmatically registered
        will try to register them again if they are "started", which will
        result in error log messages from iPOPO.
        On the other side, bundles that define factories using iPOPO decorators
        will need to be started for these factories to be registered.

        :param folder_path: The path of folder to scan
        :return: A 2-tuple, with the list of installed bundles
                 (:class:`~pelix.framework.Bundle`) and the list of the names
                 of the modules which import failed.
        :raise ValueError: Invalid path
        """

        _LOGGER.info("Loading bundles from %s", folder_path)

        bundles, failed = self.context.install_package(folder_path, True)

        for bundle in bundles:
            _LOGGER.info(
                "Installed bundle %s (ID %s )", bundle.get_symbolic_name(), bundle.get_bundle_id()
            )
            bundle.start()
        for _f in failed:
            _LOGGER.warning("Failed to import module %s", _f)

        return bundles, failed

    def get_services(
        self, service_name: str, properties: dict = None, case_sensitive: bool = False
    ) -> Optional[list]:
        """
        Returns the services that match *service_name* and provided
        *properties* (if provided).

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be
                               ignored
        :return: the list of service instances
        """
        references = self._get_service_references(service_name, properties, case_sensitive)
        services = None
        if references is not None:
            services = [self.context.get_service(ref) for ref in references]

        return services

    def register_factory(
        self,
        component_class: type,
        factory_name: str,
        service_names: Union[List[str], str],
        properties: dict = None,
    ) -> type:
        """
        Registers provided class as iPOPO component factory.

        :param component_class: the class of the components that will be provided by the factory
        :param factory_name: the name of the factory
        :param service_names: the service(s) that will be provided by the components
        :param properties: the properties associated to the factory
        :return: the input class, amended by iPOPO
        :raise FastDuplicateFactoryError:
        """
        obj = Provides(service_names)(component_class)
        with use_ipopo(self.context) as ipopo:
            if ipopo.is_registered_factory(factory_name):
                raise FastDuplicateFactoryError(factory_name)

        if properties:
            for key, value in properties.items():
                obj = Property(field="_" + self._fieldify(key), name=key, value=value)(obj)

        factory = ComponentFactory(factory_name)(obj)

        # Now factory is registered, ensure its bundle is installed and started
        self.context.install_bundle(component_class.__module__).start()

        return factory

    def get_factory_names(
        self, service_name: str = None, properties: dict = None, case_sensitive: bool = False
    ) -> List[str]:
        """
        Browses the available factory names to find what factories provide `service_name` (if
        provided) and match provided `properties` (if provided).

        if neither service_name nor properties are provided, all registered factory
        names are returned.

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be ignored
        :return: the list of factory names
        """
        with use_ipopo(self.context) as ipopo:
            all_names = ipopo.get_factories()
            if not service_name and not properties:
                return all_names

            names = []
            for name in all_names:
                details = ipopo.get_factory_details(name)
                to_be_kept = True
                if service_name and service_name not in details["services"][0]:
                    to_be_kept = False
                elif properties:
                    factory_properties: dict = details["properties"]
                    if case_sensitive:
                        to_be_kept = all(
                            item in factory_properties.items() for item in properties.items()
                        )
                    else:
                        for prop_name, prop_value in properties.items():
                            if prop_name not in factory_properties.keys():
                                to_be_kept = False
                                break
                            factory_prop_value = factory_properties[prop_name]
                            if isinstance(prop_value, str):
                                prop_value = prop_value.lower()
                                factory_prop_value = factory_prop_value.lower()
                            if prop_value != factory_prop_value:
                                to_be_kept = False
                                break
                if to_be_kept:
                    names.append(name)
        return names

    def get_factory_path(self, factory_name: str) -> str:
        """

        :param factory_name:
        :return: path of the file where the factory is defined
        """

        with use_ipopo(self.context) as ipopo:
            details = ipopo.get_factory_details(factory_name)
            bundle: Bundle = details["bundle"]
            return bundle.get_location()

    def get_factory_properties(self, factory_name: str) -> dict:
        """

        :param factory_name:
        :return: properties of the factory
        """

        with use_ipopo(self.context) as ipopo:
            details = ipopo.get_factory_details(factory_name)
            properties = details["properties"]
            return properties

    def get_factory_property(self, factory_name: str, property_name: str) -> Any:
        """

        :param factory_name:
        :param property_name:
        :return: property value, or None if property is not found
        """

        properties = self.get_factory_properties(factory_name)
        return properties.get(property_name)

    def get_instance_property(self, instance: Any, property_name: str) -> Any:
        """

        :param instance: any instance from :meth:~BundleLoader.instantiate_component
        :param property_name:
        :return: property value, or None if property is not found
        """

        try:
            return getattr(instance, "_" + self._fieldify(property_name))
        except AttributeError:
            return None

    def instantiate_component(self, factory_name: str, properties: dict = None) -> Any:
        """
        Instantiates a component from given factory

        :param factory_name: name of the factory
        :param properties: Initial properties of the component instance
        :return: the component instance
        """
        with use_ipopo(self.context) as ipopo:
            return ipopo.instantiate(
                factory_name, self._get_instance_name(factory_name), properties
            )

    def _get_instance_name(self, base_name: str):
        """
        Creates an instance name that is not currently used by iPOPO

        :param base_name: the beginning of the instance name
        :return: the created instance name
        """
        with use_ipopo(self.context) as ipopo:
            instances = ipopo.get_instances()
            instance_names = [i[0] for i in instances]
            i = 0
            name = "%s_%i" % (base_name, i)
            while name in instance_names:
                i = i + 1
                name = "%s_%i" % (base_name, i)

            return name

    def _get_service_references(
        self, service_name: str, properties: dict = None, case_sensitive: bool = False
    ) -> Optional[List[ServiceReference]]:
        """
        Returns the service references that match *service_name* and
        provided *properties* (if provided)

        :param service_name:
        :param properties:
        :param case_sensitive: if False, case of property values will be
                               ignored
        :return: a list of ServiceReference instances
        """

        # Dev Note: simple wrapper for BundleContext.get_all_service_references()

        if case_sensitive:
            operator = "="
        else:
            operator = "~="

        if not properties:
            ldap_filter = None
        else:
            ldap_filter = (
                "(&"
                + "".join(
                    [
                        "({0}{1}{2})".format(key, operator, value)
                        for key, value in properties.items()
                    ]
                )
                + ")"
            )
            _LOGGER.debug(ldap_filter)

        references = self.context.get_all_service_references(service_name, ldap_filter)
        return references

    @staticmethod
    def _fieldify(name: str) -> str:
        """
        Converts a string into a valid field name, i.e. replaces all spaces and
        non-word characters with an underscore.

        :param name: the string to fieldify
        :return: the field version of `name`
        """
        return re.compile(r"[\W_]+").sub("_", name).strip("_")



# -*- coding: utf-8 -*-
"""
    Sellar discipline 1
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .disc1_base import Disc1Base


class Disc1(Disc1Base):
    """ An OpenMDAO component to encapsulate Disc1 discipline """

    # pylint: disable=invalid-name
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Evaluates the equation
        y1 = z1**2 + z2 + x1 - 0.2*y2
        """
        z1 = inputs["z"][0]
        z2 = inputs["z"][1]
        x1 = inputs["x"]
        y2 = inputs["y2"]

        outputs["y1"] = z1 ** 2 + z2 + x1 - 0.2 * y2



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA/ISAE
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os.path as pth
from os import mkdir
from shutil import rmtree

import matplotlib.pyplot as plt
import pytest
from fastoad.io import VariableIO
from fastoad.models.performances.mission.openmdao.flight import SizingFlight
from matplotlib.ticker import MultipleLocator
from scipy.constants import foot, knot

from tests.testing_utilities import run_system

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)
    mkdir(RESULTS_FOLDER_PATH)


def plot_flight(flight_points, fig_filename):
    plt.figure(figsize=(12, 12))
    ax1 = plt.subplot(2, 1, 1)
    plt.plot(flight_points.ground_distance / 1000.0, flight_points.altitude / foot, "o-")
    plt.xlabel("distance [km]")
    plt.ylabel("altitude [ft]")
    ax1.xaxis.set_minor_locator(MultipleLocator(50))
    ax1.yaxis.set_minor_locator(MultipleLocator(500))
    plt.grid(which="major", color="k")
    plt.grid(which="minor")

    ax2 = plt.subplot(2, 1, 2)
    lines = []
    lines += plt.plot(
        flight_points.ground_distance / 1000.0, flight_points.true_airspeed, "b-", label="TAS [m/s]"
    )
    lines += plt.plot(
        flight_points.ground_distance / 1000.0,
        flight_points.equivalent_airspeed / knot,
        "g--",
        label="EAS [kt]",
    )
    plt.xlabel("distance [km]")
    plt.ylabel("speed")
    ax2.xaxis.set_minor_locator(MultipleLocator(50))
    ax2.yaxis.set_minor_locator(MultipleLocator(5))
    plt.grid(which="major", color="k")
    plt.grid(which="minor")

    plt.twinx(ax2)
    lines += plt.plot(
        flight_points.ground_distance / 1000.0, flight_points.mach, "r.-", label="Mach"
    )
    plt.ylabel("Mach")

    labels = [l.get_label() for l in lines]
    plt.legend(lines, labels, loc=0)

    plt.savefig(pth.join(RESULTS_FOLDER_PATH, fig_filename))
    plt.close()


def test_sizing_flight(cleanup):
    # problem = FASTOADProblem()
    # problem.model = SizingFlight(propulsion_id="fastoad.wrapper.propulsion.rubber_engine")
    # problem.input_file_path = pth.join(DATA_FOLDER_PATH, "flight_inputs.xml")
    # problem.write_needed_inputs(
    #     r"D:\cdavid\PyCharmProjects\FAST-OAD\tests\integration_tests\oad_process\results\non_regression\problem_outputs.xml"
    # )

    input_file_path = pth.join(DATA_FOLDER_PATH, "flight_inputs.xml")
    ivc = VariableIO(input_file_path).read().to_ivc()

    problem = run_system(
        SizingFlight(propulsion_id="fastoad.wrapper.propulsion.rubber_engine"), ivc
    )
    plot_flight(problem.model.component.flight_points, "flight.png")

    # assert_allclose(problem["data:mission:sizing:ZFW"], 65076.0, atol=1)
    # assert_allclose(problem["data:mission:sizing:fuel"], 8924.0, atol=1)



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

from fastoad.models.weight.cg.cg_components import ComputeControlSurfacesCG
from fastoad.models.weight.cg.cg_components import ComputeGlobalCG
from fastoad.models.weight.cg.cg_components import ComputeHTcg
from fastoad.models.weight.cg.cg_components import ComputeOthersCG
from fastoad.models.weight.cg.cg_components import ComputeTanksCG
from fastoad.models.weight.cg.cg_components import ComputeVTcg
from fastoad.models.weight.cg.cg_components import ComputeWingCG
from fastoad.models.weight.cg.cg_components import UpdateMLG


class CG(om.Group):
    """ Model that computes the global center of gravity """

    def setup(self):
        self.add_subsystem("ht_cg", ComputeHTcg(), promotes=["*"])
        self.add_subsystem("vt_cg", ComputeVTcg(), promotes=["*"])
        self.add_subsystem("compute_cg_wing", ComputeWingCG(), promotes=["*"])
        self.add_subsystem("compute_cg_control_surface", ComputeControlSurfacesCG(), promotes=["*"])
        self.add_subsystem("compute_cg_tanks", ComputeTanksCG(), promotes=["*"])
        self.add_subsystem("compute_cg_others", ComputeOthersCG(), promotes=["*"])
        self.add_subsystem("compute_cg", ComputeGlobalCG(), promotes=["*"])
        self.add_subsystem("update_mlg", UpdateMLG(), promotes=["*"])
        self.add_subsystem("aircraft", ComputeAircraftCG(), promotes=["*"])

        # Solvers setup
        self.nonlinear_solver = om.NonlinearBlockGS()
        self.nonlinear_solver.options["iprint"] = 0
        self.nonlinear_solver.options["maxiter"] = 200

        self.linear_solver = om.LinearBlockGS()
        self.linear_solver.options["iprint"] = 0


class ComputeAircraftCG(om.ExplicitComponent):
    """ Compute position of aircraft CG from CG ratio """

    def setup(self):
        self.add_input("data:weight:aircraft:CG:aft:MAC_position", val=np.nan)
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")

        self.add_output("data:weight:aircraft:CG:aft:x", units="m")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        cg_ratio = inputs["data:weight:aircraft:CG:aft:MAC_position"]
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        mac_position = inputs["data:geometry:wing:MAC:at25percent:x"]

        outputs["data:weight:aircraft:CG:aft:x"] = (
            mac_position - 0.25 * l0_wing + cg_ratio * l0_wing
        )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om
from fastoad.models.aerodynamics.components.compute_polar import ComputePolar, PolarType

from .components.high_lift_aero import ComputeDeltaHighLift


class AerodynamicsTakeoff(om.Group):
    """
    Computes aerodynamic characteristics at takeoff.

    - Computes CL and CD increments due to high-lift devices at takeoff.
    """

    def setup(self):
        self.add_subsystem("delta_cl_cd", ComputeDeltaHighLift(landing_flag=False), promotes=["*"])
        self.add_subsystem("polar", ComputePolar(type=PolarType.TAKEOFF), promotes=["*"])



"""
Tests for FAST-OAD variable viewer
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import pandas as pd
from pandas.util.testing import assert_frame_equal

from .. import VariableViewer

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


def test_variable_reader_display():
    """
    Basic tests for testing the VariableReader display method.
    """
    filename = pth.join(DATA_FOLDER_PATH, "problem_outputs.xml")

    # pylint: disable=invalid-name # that's a common naming
    df = VariableViewer()
    df.load(filename)

    # This is a rudimentary test as ui are difficult to verify
    # The test will fail if an error is raised by the following line
    df.display()


def test_variable_reader_load():
    """
    Basic tests for testing the VariableReader load method.
    """
    col_names = ["Name", "Value", "Unit", "Description", "I/O"]
    ref_df = pd.DataFrame()

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:seats:economical:width",
                "Value": 0.46,
                "Unit": "m",
                "Description": "width of economical class seats",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:seats:economical:length",
                "Value": 0.86,
                "Unit": "m",
                "Description": "length of economical class seats",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:aisle_width",
                "Value": 0.48,
                "Unit": "m",
                "Description": "width of aisles",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:propulsion:engine:count",
                "Value": 2.0,
                "Unit": None,
                "Description": "number of engines",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.reset_index(drop=True)

    filename = pth.join(DATA_FOLDER_PATH, "light_data.xml")

    # Testing file to df
    variable_viewer = VariableViewer()
    variable_viewer.load(filename)

    assert_frame_equal(
        ref_df.sort_values("Name").reset_index(drop=True),
        variable_viewer.dataframe.sort_values("Name").reset_index(drop=True),
    )


def test_variable_reader_save():
    """
    Basic tests for testing the VariableReader save method.
    """
    col_names = ["Name", "Value", "Unit", "Description", "I/O"]
    ref_df = pd.DataFrame()

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:seats:economical:width",
                "Value": 0.46,
                "Unit": "m",
                "Description": "width of economical class seats",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:seats:economical:length",
                "Value": 0.86,
                "Unit": "m",
                "Description": "length of economical class seats",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:cabin:aisle_width",
                "Value": 0.48,
                "Unit": "m",
                "Description": "width of aisles",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.append(
        [
            {
                "Name": "data:geometry:propulsion:engine:count",
                "Value": 2.0,
                "Unit": None,
                "Description": "number of engines",
                "I/O": "IN",
            }
        ]
    )[col_names]

    ref_df = ref_df.reset_index(drop=True)

    filename = pth.join(RESULTS_FOLDER_PATH, "light_data.xml")

    # Testing file to df
    variable_viewer = VariableViewer()
    variable_viewer.dataframe = ref_df
    variable_viewer.save(filename)

    # Loading the generated file
    variable_viewer.load(filename)

    assert_frame_equal(
        ref_df.sort_values("Name").reset_index(drop=True),
        variable_viewer.dataframe.sort_values("Name").reset_index(drop=True),
    )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ..strings import get_float_list_from_string


def test_get_float_list_from_string():
    assert [1.0, 2.0, 3.0] == get_float_list_from_string("[ 1, 2., 3]")
    assert [[1.0, 2.0], [3.0, 4.0]] == get_float_list_from_string("[[ 1, 2.],[  3, 4]]")
    assert [[1.0, 2.0], [3.0, 4.0]] == get_float_list_from_string("[[ 1,\n 2.],\n[  3, 4]]\n")
    assert [1.0, 2.0, 3.0] == get_float_list_from_string(" 1, 2., 3")
    assert [1.0, 2.0] == get_float_list_from_string(" 1 2 ")
    assert [1.0] == get_float_list_from_string(" 1 dummy 2 ")
    assert [1.0] == get_float_list_from_string(" 1     ")
    assert get_float_list_from_string(" dummy ") is None
    assert get_float_list_from_string("") is None



"""
Test module for OpenMDAO versions of RubberEngine
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.constants import EngineSetting

from tests.testing_utilities import run_system
from ..openmdao import OMRubberEngineComponent


def test_OMRubberEngineComponent():
    """ Tests ManualRubberEngine component """
    # Same test as in test_rubber_engine.test_compute_flight_points
    engine = OMRubberEngineComponent(flight_point_count=(2, 5))

    machs = [0, 0.3, 0.3, 0.8, 0.8]
    altitudes = [0, 0, 0, 10000, 13000]
    thrust_rates = [0.8, 0.5, 0.5, 0.4, 0.7]
    thrusts = [0.955 * 0.8, 0.389, 0.357, 0.0967, 0.113]
    phases = [
        EngineSetting.TAKEOFF,
        EngineSetting.TAKEOFF,
        EngineSetting.CLIMB,
        EngineSetting.IDLE,
        EngineSetting.CRUISE.value,
    ]  # mix EngineSetting with integers
    expected_sfc = [0.993e-5, 1.35e-5, 1.35e-5, 1.84e-5, 1.60e-5]

    ivc = om.IndepVarComp()
    ivc.add_output("data:propulsion:rubber_engine:bypass_ratio", 5)
    ivc.add_output("data:propulsion:rubber_engine:overall_pressure_ratio", 30)
    ivc.add_output("data:propulsion:rubber_engine:turbine_inlet_temperature", 1500, units="K")
    ivc.add_output("data:propulsion:MTO_thrust", 1, units="N")
    ivc.add_output("data:propulsion:rubber_engine:maximum_mach", 0.95)
    ivc.add_output("data:propulsion:rubber_engine:design_altitude", 10000, units="m")

    ivc.add_output("data:propulsion:mach", [machs, machs])
    ivc.add_output("data:propulsion:altitude", [altitudes, altitudes], units="m")
    ivc.add_output("data:propulsion:engine_setting", [phases, phases])
    ivc.add_output("data:propulsion:use_thrust_rate", [[True] * 5, [False] * 5])
    ivc.add_output("data:propulsion:required_thrust_rate", [thrust_rates, [0] * 5])
    ivc.add_output("data:propulsion:required_thrust", [[0] * 5, thrusts], units="N")

    problem = run_system(engine, ivc)

    np.testing.assert_allclose(
        problem["data:propulsion:SFC"], [expected_sfc, expected_sfc], rtol=1e-2
    )
    np.testing.assert_allclose(
        problem["data:propulsion:thrust_rate"], [thrust_rates, thrust_rates], rtol=1e-2
    )
    np.testing.assert_allclose(problem["data:propulsion:thrust"], [thrusts, thrusts], rtol=1e-2)



"""
Test module for rubber_engine.py
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
import pytest
from fastoad.base.flight_point import FlightPoint
from fastoad.constants import EngineSetting
from fastoad.utils.physics import Atmosphere

from ..rubber_engine import RubberEngine


def test_compute_flight_points():
    engine = RubberEngine(5, 30, 1500, 1, 0.95, 10000)  # f0=1 so that output is simply fc/f0

    # Test with scalars
    flight_point = FlightPoint(
        mach=0, altitude=0, engine_setting=EngineSetting.TAKEOFF, thrust_rate=0.8
    )  # with engine_setting as EngineSetting
    engine.compute_flight_points(flight_point)
    np.testing.assert_allclose(flight_point.thrust, 0.9553 * 0.8, rtol=1e-4)
    np.testing.assert_allclose(flight_point.sfc, 0.99210e-5, rtol=1e-4)

    flight_point = FlightPoint(
        mach=0.3, altitude=0, engine_setting=EngineSetting.CLIMB.value, thrust=0.35677
    )  # with engine_setting as int
    engine.compute_flight_points(flight_point)
    np.testing.assert_allclose(flight_point.thrust_rate, 0.5, rtol=1e-4)
    np.testing.assert_allclose(flight_point.sfc, 1.3496e-5, rtol=1e-4)

    # Test full arrays
    # 2D arrays are used, where first line is for thrust rates, and second line
    # is for thrust values.
    # As thrust rates and thrust values match, thrust rate results are 2 equal
    # lines and so are thrust value results.
    machs = [0, 0.3, 0.3, 0.8, 0.8]
    altitudes = [0, 0, 0, 10000, 13000]
    thrust_rates = [0.8, 0.5, 0.5, 0.4, 0.7]
    thrusts = [0.9553 * 0.8, 0.38851, 0.35677, 0.09680, 0.11273]
    engine_settings = [
        EngineSetting.TAKEOFF,
        EngineSetting.TAKEOFF,
        EngineSetting.CLIMB,
        EngineSetting.IDLE,
        EngineSetting.CRUISE.value,
    ]  # mix EngineSetting with integers
    expected_sfc = [0.99210e-5, 1.3496e-5, 1.3496e-5, 1.8386e-5, 1.5957e-5]

    flight_points = pd.DataFrame()
    flight_points["mach"] = machs + machs
    flight_points["altitude"] = altitudes + altitudes
    flight_points["engine_setting"] = engine_settings + engine_settings
    flight_points["thrust_is_regulated"] = [False] * 5 + [True] * 5
    flight_points["thrust_rate"] = thrust_rates + [0.0] * 5
    flight_points["thrust"] = [0.0] * 5 + thrusts
    engine.compute_flight_points(flight_points)
    np.testing.assert_allclose(flight_points.sfc, expected_sfc + expected_sfc, rtol=1e-4)
    np.testing.assert_allclose(flight_points.thrust_rate, thrust_rates + thrust_rates, rtol=1e-4)
    np.testing.assert_allclose(flight_points.thrust, thrusts + thrusts, rtol=1e-4)


def test_installed_weight():
    fj44 = RubberEngine(0, 0, 0, 8452, 0, 0)
    np.testing.assert_allclose(fj44.installed_weight(), 225, atol=1)
    br710 = RubberEngine(0, 0, 0, 66034, 0, 0)
    np.testing.assert_allclose(br710.installed_weight(), 1759, atol=1)
    cfm56_3c1 = RubberEngine(0, 0, 0, 104533, 0, 0)
    np.testing.assert_allclose(cfm56_3c1.installed_weight(), 2546, atol=1)
    trent900 = RubberEngine(0, 0, 0, 340289, 0, 0)
    np.testing.assert_allclose(trent900.installed_weight(), 6535, atol=1)


def test_length():
    engine = RubberEngine(0, 0, 0, 75000, 0.95, 0)
    np.testing.assert_allclose(engine.length(), 2.73, atol=1e-2)

    engine = RubberEngine(0, 0, 0, 250000, 0.92, 0)
    np.testing.assert_allclose(engine.length(), 4.39, atol=1e-2)


def test_nacelle_diameter():
    engine = RubberEngine(3, 0, 0, 75000, 0, 0)
    np.testing.assert_allclose(engine.nacelle_diameter(), 1.61, atol=1e-2)

    engine = RubberEngine(5.5, 0, 0, 250000, 0, 0)
    np.testing.assert_allclose(engine.nacelle_diameter(), 3.25, atol=1e-2)


def test_max_thrust():
    """
    Checks model against simplified (but analytically equivalent) formulas
    as in p. 59 of :cite:`roux:2005`, but with correct coefficients (yes, those in report
    are not consistent with the complete formula nor the figure 2.19 just below)

    .. bibliography:: ../refs.bib
    """
    engine = RubberEngine(5, 30, 1500, 1, 0, 0)  # f0=1 so that output is simply fmax/f0
    machs = np.arange(0, 1.01, 0.1)

    # Check with cruise altitude
    atm = Atmosphere(11000, altitude_in_feet=False)
    max_thrust_ratio = engine.max_thrust(atm, machs, -100)
    ref_max_thrust_ratio = (
        0.94916 * atm.density / 1.225 * (1 - 0.68060 * machs + 0.51149 * machs ** 2)
    )
    np.testing.assert_allclose(max_thrust_ratio, ref_max_thrust_ratio, rtol=1e-4)

    # Check with Takeoff altitude
    atm = Atmosphere(0, altitude_in_feet=False)
    max_thrust_ratio = engine.max_thrust(atm, machs, 0)
    ref_max_thrust_ratio = (
        0.9553 * atm.density / 1.225 * (1 - 0.72971 * machs + 0.35886 * machs ** 2)
    )
    np.testing.assert_allclose(max_thrust_ratio, ref_max_thrust_ratio, rtol=1e-4)

    # Check Cruise above 11000 with compression rate != 30 and bypass ratio != 5
    engine = RubberEngine(4, 35, 1500, 1, 0, 0)  # f0=1 so that output is simply fmax/f0
    atm = Atmosphere(13000, altitude_in_feet=False)
    max_thrust_ratio = engine.max_thrust(atm, machs, -50)
    ref_max_thrust_ratio = (
        0.96880 * atm.density / 1.225 * (1 - 0.63557 * machs + 0.52108 * machs ** 2)
    )
    np.testing.assert_allclose(max_thrust_ratio, ref_max_thrust_ratio, rtol=1e-4)

    # Check with compression rate != 30 and bypass ratio != 5 and an array for altitudes (as
    # many values as mach numbers)
    engine = RubberEngine(6, 22, 1500, 1, 0, 0)  # f0=1 so that output is simply fmax/f0
    atm = Atmosphere(np.arange(3000, 13100, 1000), altitude_in_feet=False)
    max_thrust_ratio = engine.max_thrust(atm, machs, -50)
    ref_max_thrust_ratio = [
        0.69811,
        0.59162,
        0.50117,
        0.42573,
        0.36417,
        0.31512,
        0.27704,
        0.24820,
        0.22678,
        0.19965,
        0.17795,
    ]
    np.testing.assert_allclose(max_thrust_ratio, ref_max_thrust_ratio, rtol=1e-4)


def test_sfc_at_max_thrust():
    """
    Checks model against values from :cite:`roux:2005` p.40
    (only for ground/Mach=0 values, as cruise values of the report look flawed)

    .. bibliography:: ../refs.bib
    """

    # Check with arrays
    cfm56_3c1 = RubberEngine(6, 25.7, 0, 0, 0, 0)
    atm = Atmosphere([0, 10668, 13000], altitude_in_feet=False)
    sfc = cfm56_3c1.sfc_at_max_thrust(atm, [0, 0.8, 0.8])
    # Note: value for alt==10668 is different from PhD report
    #       alt=13000 is here just for testing in stratosphere
    np.testing.assert_allclose(sfc, [0.97035e-5, 1.7756e-5, 1.7711e-5], rtol=1e-4)

    # Check with scalars
    trent900 = RubberEngine(7.14, 41, 0, 0, 0, 0)
    atm = Atmosphere(0, altitude_in_feet=False)
    sfc = trent900.sfc_at_max_thrust(atm, 0)
    np.testing.assert_allclose(sfc, 0.73469e-5, rtol=1e-4)

    atm = Atmosphere(9144, altitude_in_feet=False)
    sfc = trent900.sfc_at_max_thrust(atm, 0.8)
    np.testing.assert_allclose(sfc, 1.6766e-5, rtol=1e-4)  # value is different from PhD report

    # Check with arrays
    pw2037 = RubberEngine(6, 31.8, 0, 0, 0, 0)
    atm = Atmosphere(0, altitude_in_feet=False)
    sfc = pw2037.sfc_at_max_thrust(atm, 0)
    np.testing.assert_allclose(sfc, 0.9063e-5, rtol=1e-4)

    atm = Atmosphere(10668, altitude_in_feet=False)
    sfc = pw2037.sfc_at_max_thrust(atm, 0.85)
    np.testing.assert_allclose(sfc, 1.7439e-5, rtol=1e-4)  # value is different from PhD report


def test_sfc_ratio():
    """    Checks SFC ratio model    """
    design_alt = 10000
    engine = RubberEngine(0, 0, 0, 0, 0, design_alt)

    # Test values taken from method report (plots p. 80, see roux:2002 in refs.bib)
    # + values where original model fails (around dh=-1562.5)
    altitudes = design_alt + np.array([-2370, -1564, -1562.5, -1560, -846, 678, 2202, 3726])

    ratio = engine.sfc_ratio(altitudes, 0.8)
    assert ratio == pytest.approx(
        [1.024, 1.020, 1.020, 1.020, 1.005, 0.977, 0.948, 0.918], rel=1e-3
    )
    ratio = engine.sfc_ratio(altitudes, 0.6)
    assert ratio == pytest.approx(
        [1.074, 1.080, 1.080, 1.080, 1.044, 0.994, 0.935, 0.877], rel=1e-3
    )
    assert engine.sfc_ratio(altitudes, 1.0) == pytest.approx(1.0, rel=1e-3)

    # Because there some code differs when we have scalars:
    assert engine.sfc_ratio(design_alt - 1562.5, 0.6) == pytest.approx(1.080, rel=1e-3)



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

from .cd0_fuselage import Cd0Fuselage
from .cd0_ht import Cd0HorizontalTail
from .cd0_nacelle_pylons import Cd0NacelleAndPylons
from .cd0_total import Cd0Total
from .cd0_vt import Cd0VerticalTail
from .cd0_wing import Cd0Wing


class CD0(om.Group):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.add_subsystem(
            "cd0_wing", Cd0Wing(low_speed_aero=self.options["low_speed_aero"]), promotes=["*"]
        )
        self.add_subsystem(
            "cd0_fuselage",
            Cd0Fuselage(low_speed_aero=self.options["low_speed_aero"]),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_ht",
            Cd0HorizontalTail(low_speed_aero=self.options["low_speed_aero"]),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_vt", Cd0VerticalTail(low_speed_aero=self.options["low_speed_aero"]), promotes=["*"]
        )
        self.add_subsystem(
            "cd0_nac_pylons",
            Cd0NacelleAndPylons(low_speed_aero=self.options["low_speed_aero"]),
            promotes=["*"],
        )
        self.add_subsystem(
            "cd0_total", Cd0Total(low_speed_aero=self.options["low_speed_aero"]), promotes=["*"]
        )



"""
Test module for OpenMDAO checks
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

from .sellar_example.disc1 import Disc1
from .sellar_example.disc2 import Disc2
from .sellar_example.functions import Functions
from .sellar_example.sellar import Sellar
from ..utils import get_unconnected_input_names


def test_get_unconnected_input_names_single_component_group():
    # Test with a group problem with a single component (no promoted variable)
    group = om.Group()
    group.add_subsystem("disc1", Disc1())

    expected_mandatory_variables = {"disc1.x"}
    expected_optional_variables = {"disc1.y2", "disc1.z"}
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, False
    )
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, True
    )


def test_get_unconnected_input_names_one_component_and_ivc():
    group = om.Group()
    group.add_subsystem("disc1", Disc1())
    ivc = om.IndepVarComp()
    ivc.add_output("y2", 1.0)
    group.add_subsystem("inputs", ivc)
    group.connect("inputs.y2", "disc1.y2")

    expected_mandatory_variables = {"disc1.x"}
    expected_optional_variables = {"disc1.z"}
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, False
    )
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, True
    )


def test_get_unconnected_input_names_sellar_components():
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    group.add_subsystem("functions", Functions(), promotes=["*"])

    expected_mandatory_variables = {"disc1.x", "functions.z"}
    expected_optional_variables = {"disc1.z", "disc2.z", "functions.x"}
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, False
    )

    expected_mandatory_variables = {"z", "x"}
    expected_optional_variables = set()
    _test_problem(
        om.Problem(group), expected_mandatory_variables, expected_optional_variables, True
    )


def test_get_unconnected_input_names_full_sellar():
    expected_mandatory_variables = set()
    expected_optional_variables = set()
    _test_problem(
        om.Problem(Sellar()), expected_mandatory_variables, expected_optional_variables, False
    )
    _test_problem(
        om.Problem(Sellar()), expected_mandatory_variables, expected_optional_variables, True
    )


def _test_problem(
    problem,
    expected_missing_mandatory_variables,
    expected_missing_optional_variables,
    get_promoted_names,
):
    """ Tests get_unconnected_inputs for provided problem """

    problem.setup()

    mandatory, optional = get_unconnected_input_names(problem, promoted_names=get_promoted_names)
    assert set(mandatory) == expected_missing_mandatory_variables
    assert set(optional) == expected_missing_optional_variables



"""
Defines the analysis and plotting functions for postprocessing
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np
import plotly
import plotly.graph_objects as go
from fastoad.io import VariableIO
from plotly.subplots import make_subplots

COLS = plotly.colors.DEFAULT_PLOTLY_COLORS


def wing_geometry_plot(
    aircraft_file_path: str, name=None, fig=None, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the top view of the wing.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided, default format will
                           be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    wing_kink_leading_edge_x = variables["data:geometry:wing:kink:leading_edge:x:local"].value[0]
    wing_tip_leading_edge_x = variables["data:geometry:wing:tip:leading_edge:x:local"].value[0]
    wing_root_y = variables["data:geometry:wing:root:y"].value[0]
    wing_kink_y = variables["data:geometry:wing:kink:y"].value[0]
    wing_tip_y = variables["data:geometry:wing:tip:y"].value[0]
    wing_root_chord = variables["data:geometry:wing:root:chord"].value[0]
    wing_kink_chord = variables["data:geometry:wing:kink:chord"].value[0]
    wing_tip_chord = variables["data:geometry:wing:tip:chord"].value[0]

    # pylint: disable=invalid-name # that's a common naming
    y = np.array(
        [0, wing_root_y, wing_kink_y, wing_tip_y, wing_tip_y, wing_kink_y, wing_root_y, 0, 0]
    )
    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((-y, y))

    # pylint: disable=invalid-name # that's a common naming
    x = np.array(
        [
            0,
            0,
            wing_kink_leading_edge_x,
            wing_tip_leading_edge_x,
            wing_tip_leading_edge_x + wing_tip_chord,
            wing_kink_leading_edge_x + wing_kink_chord,
            wing_root_chord,
            wing_root_chord,
            0,
        ]
    )
    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x, x))

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=y, y=x, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig.layout = go.Layout(yaxis=dict(scaleanchor="x", scaleratio=1))

    fig = go.FigureWidget(fig)

    fig.update_layout(
        title_text="Wing Geometry", title_x=0.5, xaxis_title="y", yaxis_title="x",
    )

    return fig


# pylint: disable-msg=too-many-locals
def aircraft_geometry_plot(
    aircraft_file_path: str, name=None, fig=None, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the top view of the wing.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided, default format will
                           be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    # Wing parameters
    wing_kink_leading_edge_x = variables["data:geometry:wing:kink:leading_edge:x:local"].value[0]
    wing_tip_leading_edge_x = variables["data:geometry:wing:tip:leading_edge:x:local"].value[0]
    wing_root_y = variables["data:geometry:wing:root:y"].value[0]
    wing_kink_y = variables["data:geometry:wing:kink:y"].value[0]
    wing_tip_y = variables["data:geometry:wing:tip:y"].value[0]
    wing_root_chord = variables["data:geometry:wing:root:chord"].value[0]
    wing_kink_chord = variables["data:geometry:wing:kink:chord"].value[0]
    wing_tip_chord = variables["data:geometry:wing:tip:chord"].value[0]

    y_wing = np.array(
        [0, wing_root_y, wing_kink_y, wing_tip_y, wing_tip_y, wing_kink_y, wing_root_y, 0, 0]
    )

    x_wing = np.array(
        [
            0,
            0,
            wing_kink_leading_edge_x,
            wing_tip_leading_edge_x,
            wing_tip_leading_edge_x + wing_tip_chord,
            wing_kink_leading_edge_x + wing_kink_chord,
            wing_root_chord,
            wing_root_chord,
            0,
        ]
    )

    # Horizontal Tail parameters
    ht_root_chord = variables["data:geometry:horizontal_tail:root:chord"].value[0]
    ht_tip_chord = variables["data:geometry:horizontal_tail:tip:chord"].value[0]
    ht_span = variables["data:geometry:horizontal_tail:span"].value[0]
    ht_sweep_0 = variables["data:geometry:horizontal_tail:sweep_0"].value[0]

    ht_tip_leading_edge_x = ht_span / 2.0 * np.tan(ht_sweep_0 * np.pi / 180.0)

    y_ht = np.array([0, ht_span / 2.0, ht_span / 2.0, 0.0, 0.0])

    x_ht = np.array(
        [0, ht_tip_leading_edge_x, ht_tip_leading_edge_x + ht_tip_chord, ht_root_chord, 0]
    )

    # Fuselage parameters
    fuselage_max_width = variables["data:geometry:fuselage:maximum_width"].value[0]
    fuselage_length = variables["data:geometry:fuselage:length"].value[0]
    fuselage_front_length = variables["data:geometry:fuselage:front_length"].value[0]
    fuselage_rear_length = variables["data:geometry:fuselage:rear_length"].value[0]

    x_fuselage = np.array(
        [
            0.0,
            0.0,
            fuselage_front_length,
            fuselage_length - fuselage_rear_length,
            fuselage_length,
            fuselage_length,
        ]
    )

    y_fuselage = np.array(
        [
            0.0,
            fuselage_max_width / 4.0,
            fuselage_max_width / 2.0,
            fuselage_max_width / 2.0,
            fuselage_max_width / 4.0,
            0.0,
        ]
    )

    # CGs
    wing_25mac_x = variables["data:geometry:wing:MAC:at25percent:x"].value[0]
    wing_mac_length = variables["data:geometry:wing:MAC:length"].value[0]
    local_wing_mac_le_x = variables["data:geometry:wing:MAC:leading_edge:x:local"].value[0]
    local_ht_25mac_x = variables["data:geometry:horizontal_tail:MAC:at25percent:x:local"].value[0]
    ht_distance_from_wing = variables[
        "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25"
    ].value[0]

    x_wing = x_wing + wing_25mac_x - 0.25 * wing_mac_length - local_wing_mac_le_x
    x_ht = x_ht + wing_25mac_x + ht_distance_from_wing - local_ht_25mac_x

    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x_fuselage, x_wing, x_ht))
    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((y_fuselage, y_wing, y_ht))

    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((-y, y))
    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x, x))

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=y, y=x, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig.layout = go.Layout(yaxis=dict(scaleanchor="x", scaleratio=1))

    fig = go.FigureWidget(fig)

    fig.update_layout(
        title_text="Aircraft Geometry", title_x=0.5, xaxis_title="y", yaxis_title="x",
    )

    return fig


def drag_polar_plot(
    aircraft_file_path: str, name=None, fig=None, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the aircraft drag polar.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided, default format will
                           be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    # pylint: disable=invalid-name # that's a common naming
    cd = np.asarray(variables["data:aerodynamics:aircraft:cruise:CD"].value)
    # pylint: disable=invalid-name # that's a common naming
    cl = np.asarray(variables["data:aerodynamics:aircraft:cruise:CL"].value)

    # TODO: remove filtering one models provide proper bounds
    cd_short = cd[cd <= 2.0]
    cl_short = cl[cd <= 2.0]

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=cd_short, y=cl_short, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig = go.FigureWidget(fig)

    fig.update_layout(
        title_text="Drag Polar", title_x=0.5, xaxis_title="Cd", yaxis_title="Cl",
    )

    return fig


def mass_breakdown_bar_plot(
    aircraft_file_path: str, name=None, fig=None, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the aircraft mass breakdown using bar plots.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided, default format will
                           be assumed.
    :return: bar plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    systems = variables["data:weight:systems:mass"].value[0]

    furniture = variables["data:weight:furniture:mass"].value[0]

    crew = variables["data:weight:crew:mass"].value[0]

    airframe = variables["data:weight:airframe:mass"].value[0]

    propulsion = variables["data:weight:propulsion:mass"].value[0]

    # pylint: disable=invalid-name # that's a common naming
    MTOW = variables["data:weight:aircraft:MTOW"].value[0]
    # pylint: disable=invalid-name # that's a common naming
    OWE = variables["data:weight:aircraft:OWE"].value[0]
    payload = variables["data:weight:aircraft:payload"].value[0]
    fuel_mission = variables["data:mission:sizing:fuel"].value[0]

    if fig is None:
        fig = make_subplots(
            rows=1,
            cols=2,
            subplot_titles=("Maximum Take-Off Weight Breakdown", "Overall Weight Empty Breakdown"),
        )

    # Same color for each aircraft configuration
    i = len(fig.data)

    weight_labels = ["MTOW", "OWE", "Fuel - Mission", "Payload"]
    weight_values = [MTOW, OWE, fuel_mission, payload]
    fig.add_trace(
        go.Bar(name="", x=weight_labels, y=weight_values, marker_color=COLS[i], showlegend=False),
        row=1,
        col=1,
    )

    weight_labels = ["Airframe", "Propulsion", "Systems", "Furniture", "Crew"]
    weight_values = [airframe, propulsion, systems, furniture, crew]
    fig.add_trace(
        go.Bar(name=name, x=weight_labels, y=weight_values, marker_color=COLS[i]), row=1, col=2,
    )

    fig.update_layout(yaxis_title="[kg]")

    return fig


def mass_breakdown_sun_plot(aircraft_file_path: str, file_formatter=None):
    """
    Returns a figure sunburst plot of the mass breakdown.
    On the left a MTOW sunburst and on the right a OWE sunburst.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param file_formatter: the formatter that defines the format of data file. If not provided, default format will
                           be assumed.
    :return: sunburst plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    systems = variables["data:weight:systems:mass"].value[0]
    C11 = variables["data:weight:systems:power:auxiliary_power_unit:mass"].value[0]
    C12 = variables["data:weight:systems:power:electric_systems:mass"].value[0]
    C13 = variables["data:weight:systems:power:hydraulic_systems:mass"].value[0]
    C21 = variables["data:weight:systems:life_support:insulation:mass"].value[0]
    C22 = variables["data:weight:systems:life_support:air_conditioning:mass"].value[0]
    C23 = variables["data:weight:systems:life_support:de-icing:mass"].value[0]
    C24 = variables["data:weight:systems:life_support:cabin_lighting:mass"].value[0]
    C25 = variables["data:weight:systems:life_support:seats_crew_accommodation:mass"].value[0]
    C26 = variables["data:weight:systems:life_support:oxygen:mass"].value[0]
    C27 = variables["data:weight:systems:life_support:safety_equipment:mass"].value[0]
    C3 = variables["data:weight:systems:navigation:mass"].value[0]
    C4 = variables["data:weight:systems:transmission:mass"].value[0]
    C51 = variables["data:weight:systems:operational:radar:mass"].value[0]
    C52 = variables["data:weight:systems:operational:cargo_hold:mass"].value[0]
    C6 = variables["data:weight:systems:flight_kit:mass"].value[0]

    furniture = variables["data:weight:furniture:mass"].value[0]
    D2 = variables["data:weight:furniture:passenger_seats:mass"].value[0]
    D3 = variables["data:weight:furniture:food_water:mass"].value[0]
    D4 = variables["data:weight:furniture:security_kit:mass"].value[0]
    D5 = variables["data:weight:furniture:toilets:mass"].value[0]

    crew = variables["data:weight:crew:mass"].value[0]

    airframe = variables["data:weight:airframe:mass"].value[0]
    wing = variables["data:weight:airframe:wing:mass"].value[0]
    fuselage = variables["data:weight:airframe:fuselage:mass"].value[0]
    h_tail = variables["data:weight:airframe:horizontal_tail:mass"].value[0]
    v_tail = variables["data:weight:airframe:vertical_tail:mass"].value[0]
    control_surface = variables["data:weight:airframe:flight_controls:mass"].value[0]
    landing_gear_1 = variables["data:weight:airframe:landing_gear:main:mass"].value[0]
    landing_gear_2 = variables["data:weight:airframe:landing_gear:front:mass"].value[0]
    engine_pylon = variables["data:weight:airframe:pylon:mass"].value[0]
    paint = variables["data:weight:airframe:paint:mass"].value[0]

    propulsion = variables["data:weight:propulsion:mass"].value[0]
    B1 = variables["data:weight:propulsion:engine:mass"].value[0]
    B2 = variables["data:weight:propulsion:fuel_lines:mass"].value[0]
    B3 = variables["data:weight:propulsion:unconsumables:mass"].value[0]

    MTOW = variables["data:weight:aircraft:MTOW"].value[0]
    OWE = variables["data:weight:aircraft:OWE"].value[0]
    payload = variables["data:weight:aircraft:payload"].value[0]
    fuel_mission = variables["data:mission:sizing:fuel"].value[0]

    # TODO: Deal with this in a more generic manner ?
    if round(MTOW, 6) == round(OWE + payload + fuel_mission, 6):
        MTOW = OWE + payload + fuel_mission

    fig = make_subplots(1, 2, specs=[[{"type": "domain"}, {"type": "domain"}]],)

    fig.add_trace(
        go.Sunburst(
            labels=[
                "MTOW" + "<br>" + str(int(MTOW)) + " [kg]",
                "payload"
                + "<br>"
                + str(int(payload))
                + " [kg] ("
                + str(round(payload / MTOW * 100, 1))
                + "%)",
                "fuel_mission"
                + "<br>"
                + str(int(fuel_mission))
                + " [kg] ("
                + str(round(fuel_mission / MTOW * 100, 1))
                + "%)",
                "OWE" + "<br>" + str(int(OWE)) + " [kg] (" + str(round(OWE / MTOW * 100, 1)) + "%)",
            ],
            parents=[
                "",
                "MTOW" + "<br>" + str(int(MTOW)) + " [kg]",
                "MTOW" + "<br>" + str(int(MTOW)) + " [kg]",
                "MTOW" + "<br>" + str(int(MTOW)) + " [kg]",
            ],
            values=[MTOW, payload, fuel_mission, OWE],
            branchvalues="total",
        ),
        1,
        1,
    )

    airframe_str = (
        "airframe"
        + "<br>"
        + str(int(airframe))
        + " [kg] ("
        + str(round(airframe / OWE * 100, 1))
        + "%)"
    )
    propulsion_str = (
        "propulsion"
        + "<br>"
        + str(int(propulsion))
        + " [kg] ("
        + str(round(propulsion / MTOW * 100, 1))
        + "%)"
    )
    systems_str = (
        "systems"
        + "<br>"
        + str(int(systems))
        + " [kg] ("
        + str(round(systems / MTOW * 100, 1))
        + "%)"
    )
    furniture_str = (
        "furniture"
        + "<br>"
        + str(int(furniture))
        + " [kg] ("
        + str(round(furniture / MTOW * 100, 1))
        + "%)"
    )
    crew_str = (
        "crew" + "<br>" + str(int(crew)) + " [kg] (" + str(round(crew / MTOW * 100, 1)) + "%)"
    )

    fig.add_trace(
        go.Sunburst(
            labels=[
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                airframe_str,
                propulsion_str,
                systems_str,
                furniture_str,
                crew_str,
                "wing",
                "fuselage",
                "horizontal_tail",
                "vertical_tail",
                "flight_controls",
                "landing_gear_main",
                "landing_gear_front",
                "pylon",
                "paint",
                "engine",
                "fuel_lines",
                "unconsumables",
                "auxiliary_power_unit",
                "electric_systems",
                "hydraulic_systems",
                "insulation",
                "air_conditioning",
                "de-icing",
                "cabin_lighting",
                "seats_crew_accommodation",
                "oxygen",
                "safety_equipment",
                "navigation",
                "transmission",
                "radar",
                "cargo_hold",
                "flight_kit",
                # "cargo", "passenger_seats", "food_water", "security_kit", "toilets",
                "passenger_seats",
                "food_water",
                "security_kit",
                "toilets",
            ],
            parents=[
                "",
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                "OWE" + "<br>" + str(int(OWE)) + " [kg]",
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                airframe_str,
                propulsion_str,
                propulsion_str,
                propulsion_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                systems_str,
                # "furniture", "furniture", "furniture", "furniture", "furniture",
                furniture_str,
                furniture_str,
                furniture_str,
                furniture_str,
            ],
            values=[
                OWE,
                airframe,
                propulsion,
                systems,
                furniture,
                crew,
                wing,
                fuselage,
                h_tail,
                v_tail,
                control_surface,
                landing_gear_1,
                landing_gear_2,
                engine_pylon,
                paint,
                B1,
                B2,
                B3,
                C11,
                C12,
                C13,
                C21,
                C22,
                C23,
                C24,
                C25,
                C26,
                C27,
                C3,
                C4,
                C51,
                C52,
                C6,
                D2,
                D3,
                D4,
                D5,
            ],
            branchvalues="total",
        ),
        1,
        2,
    )

    fig.update_layout(title_text="Mass Breakdown", title_x=0.5)

    return fig



"""
Module for building OpenMDAO problem from configuration file
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth
from typing import Dict

import openmdao.api as om
import tomlkit

from fastoad.io import IVariableIOFormatter
from fastoad.models.defaults import set_all_input_defaults
from fastoad.module_management import OpenMDAOSystemRegistry
from fastoad.openmdao.problem import FASTOADProblem
from fastoad.utils.files import make_parent_dir
from .exceptions import (
    FASTConfigurationBaseKeyBuildingError,
    FASTConfigurationBadOpenMDAOInstructionError,
    FASTConfigurationError,
)

_LOGGER = logging.getLogger(__name__)  # Logger for this module

KEY_FOLDERS = "module_folders"
KEY_INPUT_FILE = "input_file"
KEY_OUTPUT_FILE = "output_file"
KEY_COMPONENT_ID = "id"
TABLE_MODEL = "model"
KEY_DRIVER = "driver"
TABLE_OPTIMIZATION = "optimization"
TABLES_DESIGN_VAR = "design_var"
TABLES_CONSTRAINT = "constraint"
TABLES_OBJECTIVE = "objective"


class FASTOADProblemConfigurator:
    """
    class for configuring an OpenMDAO problem from a configuration file

    See :ref:`description of configuration file <configuration-file>`.

    :param conf_file_path: if provided, configuration will be read directly from it
    """

    def __init__(self, conf_file_path=None):
        self._conf_file = None
        self._conf_dict = {}

        if conf_file_path:
            self.load(conf_file_path)

    @property
    def input_file_path(self):
        """path of file with input variables of the problem"""
        path = self._conf_dict[KEY_INPUT_FILE]
        if not pth.isabs(path):
            path = pth.normpath(pth.join(pth.dirname(self._conf_file), path))
        return path

    @input_file_path.setter
    def input_file_path(self, file_path: str):
        self._conf_dict[KEY_INPUT_FILE] = file_path

    @property
    def output_file_path(self):
        """path of file where output variables will be written"""
        path = self._conf_dict[KEY_OUTPUT_FILE]
        if not pth.isabs(path):
            path = pth.normpath(pth.join(pth.dirname(self._conf_file), path))
        return path

    @output_file_path.setter
    def output_file_path(self, file_path: str):
        self._conf_dict[KEY_OUTPUT_FILE] = file_path

    def get_problem(self, read_inputs: bool = False, auto_scaling: bool = False) -> FASTOADProblem:
        """
        Builds the OpenMDAO problem from current configuration.

        :param read_inputs: if True, the created problem will already be fed
                            with variables from the input file
        :param auto_scaling: if True, automatic scaling is performed for design
                             variables and constraints
        :return: the problem instance
        """
        if not self._conf_dict:
            raise RuntimeError("read configuration file first")

        problem = FASTOADProblem(self._build_model())

        problem.input_file_path = self.input_file_path
        problem.output_file_path = self.output_file_path

        driver = self._conf_dict.get(KEY_DRIVER, "")
        if driver:
            problem.driver = _om_eval(driver)

        if self.get_optimization_definition():
            self._add_constraints(problem.model, auto_scaling)
            self._add_objectives(problem.model)

        if read_inputs:
            problem.read_inputs()
            self._add_design_vars(problem.model, auto_scaling)

        return problem

    def load(self, conf_file):
        """
        Reads the problem definition

        :param conf_file: Path to the file to open or a file descriptor
        """

        self._conf_file = pth.abspath(conf_file)

        conf_dirname = pth.dirname(pth.abspath(conf_file))  # for resolving relative paths
        with open(conf_file, "r") as file:
            d = file.read()
            self._conf_dict = tomlkit.loads(d)

        # FIXME: Could structure of configuration file be checked more thoroughly ?
        for key in [KEY_INPUT_FILE, KEY_OUTPUT_FILE]:
            if key not in self._conf_dict:
                raise FASTConfigurationError(missing_key=key)

        if not isinstance(self._conf_dict.get(TABLE_MODEL), dict):
            raise FASTConfigurationError(missing_section=TABLE_MODEL)

        # Looking for modules to register
        module_folder_paths = self._conf_dict.get(KEY_FOLDERS, [])
        for folder_path in module_folder_paths:
            folder_path = pth.join(conf_dirname, folder_path)
            if not pth.exists(folder_path):
                _LOGGER.warning("SKIPPED %s: it does not exist.")
            else:
                OpenMDAOSystemRegistry.explore_folder(folder_path)

    def save(self, filename: str = None):
        """
        Saves the current configuration
        If no filename is provided, the initially read file is used.

        :param filename: file where to save configuration
        """
        if not filename:
            filename = self._conf_file

        make_parent_dir(filename)
        with open(filename, "w") as file:
            d = tomlkit.dumps(self._conf_dict)
            file.write(d)

    def write_needed_inputs(
        self, source_file_path: str = None, source_formatter: IVariableIOFormatter = None,
    ):
        """
        Writes the input file of the problem with unconnected inputs of the
        configured problem.

        Written value of each variable will be taken:

            1. from input_data if it contains the variable
            2. from defined default values in component definitions

        :param source_file_path: if provided, variable values will be read from it
        :param source_formatter: the class that defines format of input file. if
                                 not provided, expected format will be the default one.
        """
        problem = self.get_problem(read_inputs=False)
        problem.setup()
        problem.write_needed_inputs(source_file_path, source_formatter)

    def get_optimization_definition(self) -> Dict:
        """
        Returns information related to the optimization problem:
            - Design Variables
            - Constraints
            - Objectives

        :return: dict containing optimization settings for current problem
        """

        optimization_definition = {}
        conf_dict = self._conf_dict.get(TABLE_OPTIMIZATION)
        if conf_dict:
            for sec, elements in conf_dict.items():
                optimization_definition[sec] = {elem["name"]: elem for elem in elements}
        return optimization_definition

    def set_optimization_definition(self, optimization_definition: Dict):
        """
        Updates configuration with the list of design variables, constraints, objectives
        contained in the optimization_definition dictionary.

        Keys of the dictionary are: "design_var", "constraint", "objective".

        Configuration file will not be modified until :meth:`write` is used.

        :param optimization_definition: dict containing the optimization problem definition
        """
        subpart = {}
        for key, value in optimization_definition.items():
            subpart[key] = [value for _, value in optimization_definition[key].items()]
        subpart = {"optimization": subpart}
        self._conf_dict.update(subpart)

    def _build_model(self) -> om.Group:
        """
        Builds the model as defined in the configuration file.

        The model is initialized as a new group and populated with subsystems
        indicated in configuration file.

        :return: the built model
        """

        model = om.Group()
        model_definition = self._conf_dict.get(TABLE_MODEL)

        try:
            if KEY_COMPONENT_ID in model_definition:
                # The defined model is only one system
                system_id = model_definition[KEY_COMPONENT_ID]
                sub_component = OpenMDAOSystemRegistry.get_system(system_id)
                model.add_subsystem("system", sub_component, promotes=["*"])
            else:
                # The defined model is a group
                self._parse_problem_table(model, model_definition)

        except FASTConfigurationBaseKeyBuildingError as err:
            log_err = err.__class__(err, TABLE_MODEL)
            _LOGGER.error(log_err)
            raise log_err

        set_all_input_defaults(model)
        return model

    def _parse_problem_table(self, group: om.Group, table: dict):
        """
        Feeds provided *group*, using definition in provided TOML *table*.

        :param group:
        :param table:
        """
        assert isinstance(table, dict), "table should be a dictionary"

        for key, value in table.items():
            if isinstance(value, dict):  # value defines a sub-component
                if KEY_COMPONENT_ID in value:
                    # It is a non-group component, that should be registered with its ID
                    options = value.copy()
                    identifier = options.pop(KEY_COMPONENT_ID)
                    sub_component = OpenMDAOSystemRegistry.get_system(identifier, options=options)
                    group.add_subsystem(key, sub_component, promotes=["*"])
                else:
                    # It is a Group
                    sub_component = group.add_subsystem(key, om.Group(), promotes=["*"])
                    try:
                        self._parse_problem_table(sub_component, value)
                    except FASTConfigurationBadOpenMDAOInstructionError as err:
                        # There has been an error while parsing an attribute.
                        # Error is relayed with key added for context
                        raise FASTConfigurationBadOpenMDAOInstructionError(err, key)
            else:
                # value is an attribute of current component and will be literally interpreted
                try:
                    setattr(group, key, _om_eval(value))  # pylint:disable=eval-used
                except Exception as err:
                    raise FASTConfigurationBadOpenMDAOInstructionError(err, key, value)

    def _add_constraints(self, model, auto_scaling):
        """
        Adds constraints to provided model as instructed in current configuration

        :param model:
        :param auto_scaling:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        # Constraints
        constraint_tables = optimization_definition.get(TABLES_CONSTRAINT, {})
        for constraint_table in constraint_tables.values():
            if (
                auto_scaling
                and "lower" in constraint_table
                and "upper" in constraint_table
                and constraint_table.get("ref0") is not None
                and constraint_table.get("ref") is not None
                and constraint_table["lower"] != constraint_table["upper"]
            ):
                constraint_table["ref0"] = constraint_table["lower"]
                constraint_table["ref"] = constraint_table["upper"]
            model.add_constraint(**constraint_table)

    def _add_objectives(self, model):
        """
        Adds objectives to provided model as instructed in current configuration

        :param model:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        objective_tables = optimization_definition.get(TABLES_OBJECTIVE, {})
        for objective_table in objective_tables.values():
            model.add_objective(**objective_table)

    def _add_design_vars(self, model, auto_scaling):
        """
        Adds design variables to provided model as instructed in current configuration

        :param model:
        :param auto_scaling:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        design_var_tables = optimization_definition.get(TABLES_DESIGN_VAR, {})
        for design_var_table in design_var_tables.values():
            if (
                auto_scaling
                and "lower" in design_var_table
                and "upper" in design_var_table
                and design_var_table.get("ref0") is not None
                and design_var_table.get("ref") is not None
                and design_var_table["lower"] != design_var_table["upper"]
            ):
                design_var_table["ref0"] = design_var_table["lower"]
                design_var_table["ref"] = design_var_table["upper"]
            model.add_design_var(**design_var_table)


def _om_eval(string_to_eval: str):
    """
    Evaluates strings that assume `import openmdao.api as om` is done.

    eval() is used for that, as safely as possible.

    :param string_to_eval:
    :return: result of eval()
    """
    if "__" in string_to_eval:
        raise ValueError("No double underscore allowed in evaluated string for security reasons")
    return eval(string_to_eval, {"__builtins__": {}}, {"om": om})



"""
Estimation of passenger seats weight
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

from fastoad.constants import RangeCategory


class PassengerSeatsWeight(om.ExplicitComponent):
    """
    Weight estimation for passenger seats

    Based on :cite:`supaero:2014`, mass contribution D2
    """

    def setup(self):
        self.add_input("data:TLAR:range", val=np.nan, units="NM")
        self.add_input("data:TLAR:NPAX", val=np.nan)
        self.add_input("tuning:weight:furniture:passenger_seats:mass:k", val=1.0)
        self.add_input("tuning:weight:furniture:passenger_seats:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:furniture:passenger_seats:mass", units="kg")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        tlar_range = inputs["data:TLAR:range"]
        npax = inputs["data:TLAR:NPAX"]
        k_d2 = inputs["tuning:weight:furniture:passenger_seats:mass:k"]
        offset_d2 = inputs["tuning:weight:furniture:passenger_seats:mass:offset"]

        if tlar_range in RangeCategory.SHORT:
            k_ps = 9.0
        elif RangeCategory.SHORT_MEDIUM.min() <= tlar_range <= RangeCategory.MEDIUM.max():
            k_ps = 10.0
        else:
            k_ps = 11.0

        temp_d2 = k_ps * npax
        outputs["data:weight:furniture:passenger_seats:mass"] = k_d2 * temp_d2 + offset_d2



"""Base classes for fuel-consuming propulsion models."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC
from typing import Union

import pandas as pd

from fastoad.base.flight_point import FlightPoint
from fastoad.models.propulsion import IPropulsion


class AbstractFuelPropulsion(IPropulsion, ABC):
    """
    Propulsion model that consume any fuel should inherit from this one.

    In inheritors, :meth:`compute_flight_points` is expected to define
    "sfc" and "thrust" in computed FlightPoint instances.
    """

    def get_consumed_mass(self, flight_point: FlightPoint, time_step: float) -> float:
        return time_step * flight_point.sfc * flight_point.thrust


class FuelEngineSet(AbstractFuelPropulsion):
    def __init__(self, engine: IPropulsion, engine_count):
        """
        Class for modelling an assembly of identical fuel engines.

        Thrust is supposed equally distributed among them.

        :param engine: the engine model
        :param engine_count:
        """
        self.engine = engine
        self.engine_count = engine_count

    def compute_flight_points(self, flight_points: Union[FlightPoint, pd.DataFrame]):

        if isinstance(flight_points, FlightPoint):
            flight_points_per_engine = FlightPoint(flight_points)
        else:
            flight_points_per_engine = flight_points.copy()

        if flight_points.thrust is not None:
            flight_points_per_engine.thrust = flight_points.thrust / self.engine_count

        self.engine.compute_flight_points(flight_points_per_engine)
        flight_points.sfc = flight_points_per_engine.sfc
        flight_points.thrust = flight_points_per_engine.thrust * self.engine_count
        flight_points.thrust_rate = flight_points_per_engine.thrust_rate



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from filecmp import dircmp
from glob import glob
from shutil import rmtree

import pytest

from . import resources
from ..copy import copy_resource_folder

RESOURCE_FOLDER_PATH = pth.join(pth.dirname(__file__), "resources")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def test_copy_resource_folder_from_str(cleanup):
    destination_folder = pth.join(RESULTS_FOLDER_PATH, "test_from_str")
    copy_resource_folder("fastoad.utils.resource_management.tests.resources", destination_folder)
    _check_files(destination_folder)


def test_copy_resource_folder_from_str_with_exclusion(cleanup):
    destination_folder = pth.join(RESULTS_FOLDER_PATH, "test_from_str_with_exclusion")
    excluded = ["__init__.py"]
    copy_resource_folder(
        "fastoad.utils.resource_management.tests.resources", destination_folder, exclude=excluded,
    )
    _check_files(destination_folder, excluded)


def test_copy_resource_folder_from_package(cleanup):
    destination_folder = pth.join(RESULTS_FOLDER_PATH, "test_from_module")
    copy_resource_folder(resources, destination_folder)
    _check_files(destination_folder)


def test_copy_resource_folder_from_package_with_exclusion(cleanup):
    destination_folder = pth.join(RESULTS_FOLDER_PATH, "test_from_str_with_exclusion")
    excluded = ["__init__.py"]
    copy_resource_folder(resources, destination_folder, exclude=excluded)
    _check_files(destination_folder, excluded)


def _check_files(destination_folder, excluded=None):
    assert dircmp(RESOURCE_FOLDER_PATH, destination_folder, ignore=excluded)

    if excluded:
        for name in excluded:
            assert not glob(pth.join(destination_folder, "**", name), recursive=True)
